# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from typing import Any, Callable, Optional, Sequence, Tuple, TypeVar, Union

import jax
from jax import lax
from jax import random
from jax.tree_util import Partial, register_pytree_node_class

from . import conjugate_gradient
from .forest_util import assert_arithmetics, map_forest, map_forest_mean, unstack
from .likelihood import Likelihood, StandardHamiltonian
from .sugar import random_like

P = TypeVar("P")


def sample_likelihood(likelihood: Likelihood, primals, key):
    white_sample = random_like(key, likelihood.left_sqrt_metric_tangents_shape)
    return likelihood.left_sqrt_metric(primals, white_sample)


def cond_raise(condition, exception):
    from jax.experimental.host_callback import call

    def maybe_raise(condition):
        if condition:
            raise exception

    call(maybe_raise, condition, result_shape=None)


# TODO (?): optionally accept ham.metric and likelihood.lsm
def _sample_standard_hamiltonian(
    hamiltonian: StandardHamiltonian,
    primals,
    key,
    from_inverse: bool,
    cg: Callable = conjugate_gradient.static_cg,
    cg_name: Optional[str] = None,
    cg_kwargs: Optional[dict] = None,
):
    if not isinstance(hamiltonian, StandardHamiltonian):
        te = f"`hamiltonian` of invalid type; got '{type(hamiltonian)}'"
        raise TypeError(te)
    cg_kwargs = cg_kwargs if cg_kwargs is not None else {}

    subkey_nll, subkey_prr = random.split(key, 2)
    nll_smpl = sample_likelihood(
        hamiltonian.likelihood, primals, key=subkey_nll
    )
    prr_inv_metric_smpl = random_like(key=subkey_prr, primals=primals)
    # One may transform any metric sample to a sample of the inverse
    # metric by simply applying the inverse metric to it
    prr_smpl = prr_inv_metric_smpl
    # Note, we can sample antithetically by swapping the global sign of
    # the metric sample below (which corresponds to mirroring the final
    # sample) and additionally by swapping the relative sign between
    # the prior and the likelihood sample. The first technique is
    # computationally cheap and empirically known to improve stability.
    # The latter technique requires an additional inversion and its
    # impact on stability is still unknown.
    # TODO: investigate the impact of sampling the prior and likelihood
    # antithetically.
    met_smpl = nll_smpl + prr_smpl
    if from_inverse:
        inv_metric_at_p = partial(
            cg, Partial(hamiltonian.metric, primals), **{
                "name": cg_name,
                **cg_kwargs
            }
        )
        signal_smpl, info = inv_metric_at_p(met_smpl, x0=prr_inv_metric_smpl)
        cond_raise(
            (info is not None) & (info < 0),
            ValueError("conjugate gradient failed")
        )
        return signal_smpl, met_smpl
    else:
        return None, met_smpl


def sample_standard_hamiltonian(
    hamiltonian: StandardHamiltonian, primals, *args, **kwargs
):
    r"""Draws a sample of which the covariance is the metric or the inverse
    metric of the Hamiltonian.

    To sample from the inverse metric, we need to be able to draw samples
    which have the metric as covariance structure and we need to be able to
    apply the inverse metric. The first part is trivial since we can use
    the left square root of the metric :math:`L` associated with every
    likelihood:

    .. math::

        \tilde{d} \leftarrow \mathcal{G}(0,\mathbb{1}) \\
        t = L \tilde{d}

    with :math:`t` now having a covariance structure of

    .. math::
        <t t^\dagger> = L <\tilde{d} \tilde{d}^\dagger> L^\dagger = M .

    We now need to apply the inverse metric in order to transform the
    sample to an inverse sample. We can do so using the conjugate gradient
    algorithm which yields the solution to :math:`M s = t`, i.e. applies the
    inverse of :math:`M` to :math:`t`:

    .. math::

        M &s =  t \\
        &s = M^{-1} t = cg(M, t) .

    Parameters
    ----------
    hamiltonian:
        Hamiltonian with standard prior from which to draw samples.
    primals : tree-like structure
        Position at which to draw samples.
    key : tuple, list or jnp.ndarray of uint32 of length two
        Random key with which to generate random variables in data domain.
    cg : callable, optional
        Implementation of the conjugate gradient algorithm and used to
        apply the inverse of the metric.
    cg_kwargs : dict, optional
        Additional keyword arguments passed on to `cg`.

    Returns
    -------
    sample : tree-like structure
        Sample of which the covariance is the inverse metric.

    See also
    --------
    `Metric Gaussian Variational Inference`, Jakob Knollmüller,
    Torsten A. Enßlin, `<https://arxiv.org/abs/1901.11033>`_
    """
    assert_arithmetics(primals)
    inv_met_smpl, _ = _sample_standard_hamiltonian(
        hamiltonian, primals, *args, from_inverse=True, **kwargs
    )
    return inv_met_smpl


# TODO (?): optionally accept ham.metric and likelihood.lsm and likelihood.transformation
def geometrically_sample_standard_hamiltonian(
    hamiltonian: StandardHamiltonian,
    primals,
    key,
    mirror_linear_sample: bool,
    linear_sampling_cg: Callable = conjugate_gradient.static_cg,
    linear_sampling_name: Optional[str] = None,
    linear_sampling_kwargs: Optional[dict] = None,
    non_linear_sampling_method: str = "NewtonCG",
    non_linear_sampling_name: Optional[str] = None,
    non_linear_sampling_kwargs: Optional[dict] = None,
):
    r"""Draws a sample which follows a standard normal distribution in the
    canonical coordinate system of the Riemannian manifold associated with the
    metric of the other distribution. The coordinate transformation is
    approximated by expanding around a given point `primals`.

    Parameters
    ----------
    hamiltonian:
        Hamiltonian with standard prior from which to draw samples.
    primals : tree-like structure
        Position at which to draw samples.
    key : tuple, list or jnp.ndarray of uint32 of length two
        Random key with which to generate random variables in data domain.
    linear_sampling_cg : callable
        Implementation of the conjugate gradient algorithm and used to
        apply the inverse of the metric.
    linear_sampling_kwargs : dict
        Additional keyword arguments passed on to `cg`.
    non_linear_sampling_kwargs : dict
        Additional keyword arguments passed on to the minimzer of the
        non-linear potential.

    Returns
    -------
    sample : tree-like structure
        Sample of which the covariance is the inverse metric.

    See also
    --------
    `Geometric Variational Inference`, Philipp Frank, Reimar Leike,
    Torsten A. Enßlin, `<https://arxiv.org/abs/2105.10470>`_
    `<https://doi.org/10.3390/e23070853>`_
    """
    if not isinstance(hamiltonian, StandardHamiltonian):
        te = f"`hamiltonian` of invalid type; got '{type(hamiltonian)}'"
        raise TypeError(te)
    assert_arithmetics(primals)
    from .energy_operators import Gaussian
    from .optimize import minimize

    inv_met_smpl, met_smpl = _sample_standard_hamiltonian(
        hamiltonian,
        primals,
        key=key,
        from_inverse=True,
        cg=linear_sampling_cg,
        cg_name=linear_sampling_name,
        cg_kwargs=linear_sampling_kwargs
    )

    if isinstance(non_linear_sampling_kwargs, dict):
        nls_kwargs = non_linear_sampling_kwargs
    elif non_linear_sampling_kwargs is None:
        nls_kwargs = {}
    else:
        te = (
            "`non_linear_sampling_kwargs` of invalid type"
            "{type(non_linear_sampling_kwargs)}"
        )
        raise TypeError(te)
    nls_kwargs = {"name": non_linear_sampling_name, **nls_kwargs}
    if "hessp" in nls_kwargs:
        ve = "setting the hessian for an unknown function is invalid"
        raise ValueError(ve)
    # Abort early if non-linear sampling is effectively disabled
    if nls_kwargs.get("maxiter") == 0:
        if mirror_linear_sample:
            return (inv_met_smpl, -inv_met_smpl)
        return (inv_met_smpl, )

    lh_trafo_at_p = hamiltonian.likelihood.transformation(primals)

    def draw_non_linear_sample(lh, met_smpl, inv_met_smpl):
        x0 = primals + inv_met_smpl

        def g(x):
            return x - primals + lh.left_sqrt_metric(
                primals,
                lh.transformation(x) - lh_trafo_at_p
            )

        r2_half = Gaussian(met_smpl) @ g  # (g - met_smpl)**2 / 2

        options = nls_kwargs.copy()
        options["hessp"] = r2_half.metric

        opt_state = minimize(
            r2_half, x0=x0, method=non_linear_sampling_method, options=options
        )

        return opt_state.x, opt_state.status

    smpl1, smpl1_status = draw_non_linear_sample(
        hamiltonian.likelihood, met_smpl, inv_met_smpl
    )
    cond_raise(
        (smpl1_status is not None) & (smpl1_status < 0),
        ValueError("S: failed to invert map")
    )
    if not mirror_linear_sample:
        return (smpl1 - primals, )
    smpl2, smpl2_status = draw_non_linear_sample(
        hamiltonian.likelihood, -met_smpl, -inv_met_smpl
    )
    cond_raise(
        (smpl2_status is not None) & (smpl2_status < 0),
        ValueError("S: failed to invert map")
    )
    return (smpl1 - primals, smpl2 - primals)


@register_pytree_node_class
class SampleIter():
    """Storage class for samples with some convenience methods for applying
    operators of them

    This class is used to store samples for the Variational Inference schemes
    MGVI and geoVI where samples are defined relative to some expansion point
    (a.k.a. latent mean or offset).

    See also
    --------
    `Geometric Variational Inference`, Philipp Frank, Reimar Leike,
    Torsten A. Enßlin, `<https://arxiv.org/abs/2105.10470>`_
    `<https://doi.org/10.3390/e23070853>`_

    `Metric Gaussian Variational Inference`, Jakob Knollmüller,
    Torsten A. Enßlin, `<https://arxiv.org/abs/1901.11033>`_
    """
    def __init__(
        self,
        *,
        mean: P = None,
        samples: Sequence[P],
        linearly_mirror_samples: bool = False,
    ):
        self._samples = tuple(samples)
        self._mean = mean

        self._n_samples = len(self._samples)
        if linearly_mirror_samples == True:
            self._n_samples *= 2
        self._linearly_mirror_samples = linearly_mirror_samples
        # TODO/IDEA: Implement a transposed SampleIter object (SampleStack)
        # akin to `vmap_forest_mean`

    def __iter__(self):
        for s in self._samples:
            yield self._mean + s if self._mean is not None else s
            if self._linearly_mirror_samples:
                yield self._mean - s if self._mean is not None else -s

    def __len__(self):
        return self._n_samples

    @property
    def n_samples(self):
        """Total number of samples, equivalent to the length of the object"""
        return len(self)

    def at(self, mean):
        """Updates the offset (usually the latent mean) of all samples"""
        return SampleIter(
            mean=mean,
            samples=self._samples,
            linearly_mirror_samples=self._linearly_mirror_samples
        )

    @property
    def first(self):
        """Convenience method to easily retrieve a sample (the first one)"""
        if self._mean is not None:
            return self._mean + self._samples[0]
        return self._samples[0]

    def apply(self, call: Callable, *args, **kwargs):
        """Applies an operator over all samples, yielding a list of outputs

        Internally, the call is `vmap`-ed over the samples for additional
        efficiency.
        """
        if set(kwargs.keys()) | {"in_axes"} != {"in_axes"}:
            raise ValueError(f"invalid keyword arguments {kwargs}")

        # TODO: vmap is significantly slower than looping over the samples
        # for an extremely high dimensional problem.
        in_axes = kwargs.get("in_axes", (0, ))
        return map_forest(call, in_axes=in_axes)(tuple(self), *args)

    def mean(self, call: Callable, *args, **kwargs):
        """Applies an operator over all samples and averages the results

        Internally, the call is `vmap`-ed over the samples for additional
        efficiency.
        """
        if set(kwargs.keys()) | {"in_axes"} != {"in_axes"}:
            raise ValueError(f"invalid keyword arguments {kwargs}")

        # TODO: vmap is significantly slower than looping over the samples
        # for an extremely high dimensional problem.
        in_axes = kwargs.get("in_axes", (0, ))
        return map_forest_mean(call, in_axes=in_axes)(tuple(self), *args)

    def tree_flatten(self):
        return ((self._mean, self._samples), (self._linearly_mirror_samples, ))

    @classmethod
    def tree_unflatten(cls, aux, children):
        if len(aux) != 1 or len(children) != 2:
            raise ValueError()
        return cls(
            mean=children[0],
            samples=children[1],
            linearly_mirror_samples=aux[0]
        )


def MetricKL(
    hamiltonian: StandardHamiltonian,
    primals,
    n_samples: int,
    key,
    mirror_samples: bool = True,
    sample_mapping: Union[str, Callable] = 'lax',
    linear_sampling_cg: Callable = conjugate_gradient.static_cg,
    linear_sampling_name: Optional[str] = None,
    linear_sampling_kwargs: Optional[dict] = None,
) -> SampleIter:
    """Provides the sampled Kullback-Leibler divergence between a distribution
    and a Metric Gaussian.

    A Metric Gaussian is used to approximate another probability distribution.
    It is a Gaussian distribution that uses the Fisher information metric of
    the other distribution at the location of its mean to approximate the
    variance. In order to infer the mean, a stochastic estimate of the
    Kullback-Leibler divergence is minimized. This estimate is obtained by
    sampling the Metric Gaussian at the current mean. During minimization these
    samples are kept constant and only the mean is updated. Due to the
    typically nonlinear structure of the true distribution these samples have
    to be updated eventually by re-instantiating the Metric Gaussian again. For
    the true probability distribution the standard parametrization is assumed.

    Parameters
    ----------
    hamiltonian : :class:`nifty8.src.re.likelihood.StandardHamiltonian`
        Hamiltonian of the approximated probability distribution.
    primals : :class:`nifty8.re.field.Field`
        Expansion point of the coordinate transformation.
    n_samples : integer
        Number of samples used to stochastically estimate the KL.
    key : DeviceArray
        A PRNG-key.
    mirror_samples : bool
        Whether the mirrored version of the drawn samples are also used.
        If true, the number of used samples doubles.
        Mirroring samples stabilizes the KL estimate as extreme
        sample variation is counterbalanced.
        Default is True.
    sample_mapping : string, callable
        Can be either a string-key to a mapping function or a mapping function
        itself. The function is used to map the drawing of samples. Possible
        string-keys are:

        - 'pmap' or 'p' for `jax.pmap`
        - 'lax.map' or 'lax' for `jax.lax.map`

        In case sample_mapping is passed as a function, it should produce a
        mapped function f_mapped of a general function f as: `f_mapped =
        sample_mapping(f)`
    linear_sampling_cg : callable
        Implementation of the conjugate gradient algorithm and used to
        apply the inverse of the metric.
    linear_sampling_name : string, optional
        'name'-keyword-argument passed to `linear_sampling_cg`.
    linear_sampling_kwargs : dict, optional
        Additional keyword arguments passed on to `linear_sampling_cg`.

    See also
    --------
    `Metric Gaussian Variational Inference`, Jakob Knollmüller,
    Torsten A. Enßlin, `<https://arxiv.org/abs/1901.11033>`_
    """
    if not isinstance(hamiltonian, StandardHamiltonian):
        te = f"`hamiltonian` of invalid type; got '{type(hamiltonian)}'"
        raise TypeError(te)
    assert_arithmetics(primals)

    draw = partial(
        sample_standard_hamiltonian,
        hamiltonian=hamiltonian,
        primals=primals,
        cg=linear_sampling_cg,
        cg_name=linear_sampling_name,
        cg_kwargs=linear_sampling_kwargs
    )
    subkeys = random.split(key, n_samples)
    if isinstance(sample_mapping, str):
        if sample_mapping == 'pmap' or sample_mapping == 'p':
            sample_mapping = jax.pmap
        elif sample_mapping == 'lax.map' or sample_mapping == 'lax':
            sample_mapping = partial(partial, lax.map)
        else:
            ve = (
                f"{sample_mapping} is not an accepted key to a mapping function"
                "; please pass function directly"
            )
            raise ValueError(ve)

    elif not callable(sample_mapping):
        te = (
            f"invalid `sample_mapping` of type {type(sample_mapping)!r}"
            "; expected string or callable"
        )
        raise TypeError(te)

    samples_stack = sample_mapping(lambda k: draw(key=k))(subkeys)

    return SampleIter(
        mean=primals,
        samples=unstack(samples_stack),
        linearly_mirror_samples=mirror_samples
    )


def GeoMetricKL(
    hamiltonian: StandardHamiltonian,
    primals,
    n_samples: int,
    key,
    mirror_samples: bool = True,
    linear_sampling_cg: Callable = conjugate_gradient.static_cg,
    linear_sampling_name: Optional[str] = None,
    linear_sampling_kwargs: Optional[dict] = None,
    non_linear_sampling_method: str = "NewtonCG",
    non_linear_sampling_name: Optional[str] = None,
    non_linear_sampling_kwargs: Optional[dict] = None,
) -> SampleIter:
    """Provides the sampled Kullback-Leibler used in geometric Variational
    Inference (geoVI).

    In geoVI a probability distribution is approximated with a standard normal
    distribution in the canonical coordinate system of the Riemannian manifold
    associated with the metric of the other distribution. The coordinate
    transformation is approximated by expanding around a point. In order to
    infer the expansion point, a stochastic estimate of the Kullback-Leibler
    divergence is minimized. This estimate is obtained by sampling from the
    approximation using the current expansion point. During minimization these
    samples are kept constant and only the expansion point is updated. Due to
    the typically nonlinear structure of the true distribution these samples
    have to be updated eventually by re-instantiating the geometric Gaussian
    again. For the true probability distribution the standard parametrization
    is assumed.

    See also
    --------
    `Geometric Variational Inference`, Philipp Frank, Reimar Leike,
    Torsten A. Enßlin, `<https://arxiv.org/abs/2105.10470>`_
    `<https://doi.org/10.3390/e23070853>`_
    """
    if not isinstance(hamiltonian, StandardHamiltonian):
        te = f"`hamiltonian` of invalid type; got '{type(hamiltonian)}'"
        raise TypeError(te)
    assert_arithmetics(primals)

    draw = partial(
        geometrically_sample_standard_hamiltonian,
        hamiltonian=hamiltonian,
        primals=primals,
        mirror_linear_sample=mirror_samples,
        linear_sampling_cg=linear_sampling_cg,
        linear_sampling_name=linear_sampling_name,
        linear_sampling_kwargs=linear_sampling_kwargs,
        non_linear_sampling_method=non_linear_sampling_method,
        non_linear_sampling_name=non_linear_sampling_name,
        non_linear_sampling_kwargs=non_linear_sampling_kwargs
    )
    subkeys = random.split(key, n_samples)
    # TODO: Make `geometrically_sample_standard_hamiltonian` jit-able
    # samples_stack = lax.map(lambda k: draw(key=k), subkeys)
    # Unpack tuple of samples
    # samples_stack = tree_map(
    #     lambda a: a.reshape((-1, ) + a.shape[2:]), samples_stack
    # )
    # samples = unstack(samples_stack)
    samples = tuple(s for ss in map(lambda k: draw(key=k), subkeys) for s in ss)

    return SampleIter(
        mean=primals, samples=samples, linearly_mirror_samples=False
    )


def mean_value_and_grad(ham: Callable, sample_mapping='vmap', *args, **kwargs):
    """Thin wrapper around `value_and_grad` and the provided sample mapping
    function, e.g. `vmap` to apply a cost function to a mean and a list of
    residual samples.

    Parameters
    ----------
    ham : :class:`nifty8.src.re.likelihood.StandardHamiltonian`
        Hamiltonian of the approximated probability distribution,
        of which the mean value and the mean gradient are to be computed.
    sample_mapping : string, callable
        Can be either a string-key to a mapping function or a mapping function
        itself. The function is used to map the drawing of samples. Possible
        string-keys are:

        - 'vmap' or 'v' for `jax.vmap`
        - 'pmap' or 'p' for `jax.pmap`
        - 'lax.map' or 'lax' for `jax.lax.map`

        In case sample_mapping is passed as a function, it should produce a
        mapped function f_mapped of a general function f as: `f_mapped =
        sample_mapping(f)`
    """
    from jax import value_and_grad
    vg = value_and_grad(ham, *args, **kwargs)

    def mean_vg(
        primals: P,
        primals_samples: Union[None, Sequence[P], SampleIter] = None,
        **primals_kw
    ) -> Tuple[Any, P]:
        ham_vg = partial(vg, **primals_kw)
        if primals_samples is None:
            return ham_vg(primals)

        if not isinstance(primals_samples, SampleIter):
            primals_samples = SampleIter(samples=primals_samples)
        return map_forest_mean(ham_vg, mapping=sample_mapping, in_axes=(0, ))(
            tuple(primals_samples.at(primals))
        )

    return mean_vg


def mean_hessp(ham: Callable, *args, **kwargs):
    """Thin wrapper around `jvp`, `grad` and `vmap` to apply a binary method to
    a primal mean, a tangent and a list of residual primal samples.
    """
    from jax import jvp, grad
    jac = grad(ham, *args, **kwargs)

    def mean_hp(
        primals: P,
        tangents: Any,
        primals_samples: Union[None, Sequence[P], SampleIter] = None,
        **primals_kw
    ) -> P:
        if primals_samples is None:
            _, hp = jvp(partial(jac, **primals_kw), (primals, ), (tangents, ))
            return hp

        if not isinstance(primals_samples, SampleIter):
            primals_samples = SampleIter(samples=primals_samples)
        return map_forest_mean(
            partial(mean_hp, primals_samples=None, **primals_kw),
            in_axes=(0, None)
        )(tuple(primals_samples.at(primals)), tangents)

    return mean_hp


def mean_metric(metric: Callable):
    """Thin wrapper around `vmap` to apply a binary method to a primal mean, a
    tangent and a list of residual primal samples.
    """
    def mean_met(
        primals: P,
        tangents: Any,
        primals_samples: Union[None, Sequence[P], SampleIter] = None,
        **primals_kw
    ) -> P:
        if primals_samples is None:
            return metric(primals, tangents, **primals_kw)

        if not isinstance(primals_samples, SampleIter):
            primals_samples = SampleIter(samples=primals_samples)
        return map_forest_mean(
            partial(metric, **primals_kw), in_axes=(0, None)
        )(tuple(primals_samples.at(primals)), tangents)

    return mean_met



# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from typing import Any, Callable, NamedTuple, Optional, Tuple, Union

import numpy as np
from jax import grad
from jax import numpy as jnp
from jax import random, tree_util

from .disable_jax_control_flow import fori_loop
from .hmc import AcceptedAndRejected, Q, QP, Tree
from .hmc import (
    generate_hmc_acc_rej,
    generate_nuts_tree,
    leapfrog_step,
    sample_momentum_from_diagonal,
    tree_index_update,
)


def _parse_diag_mass_matrix(mass_matrix, position_proto: Q) -> Q:
    if isinstance(mass_matrix,
                  (float, jnp.ndarray)) and jnp.size(mass_matrix) == 1:
        mass_matrix = tree_util.tree_map(
            partial(jnp.full_like, fill_value=mass_matrix), position_proto
        )
    elif tree_util.tree_structure(mass_matrix
                                 ) == tree_util.tree_structure(position_proto):
        shape_match_tree = tree_util.tree_map(
            lambda a1, a2: jnp.shape(a1) == jnp.shape(a2), mass_matrix,
            position_proto
        )
        shape_and_structure_match = all(
            tree_util.tree_flatten(shape_match_tree)
        )
        if not shape_and_structure_match:
            ve = "matrix has same tree_structe as the position but shapes do not match up"
            raise ValueError(ve)
    else:
        te = "matrix must either be float or have same tree structure as the position"
        raise TypeError(te)

    return mass_matrix


class Chain(NamedTuple):
    """Object carrying chain metadata; think: transposed Tree with new axis.
    """
    # Q but with one more dimension on the first axes of the leave tensors
    samples: Q
    divergences: jnp.ndarray
    acceptance: Union[jnp.ndarray, float]
    depths: Optional[jnp.ndarray] = None
    trees: Optional[Union[Tree, AcceptedAndRejected]] = None


class _Sampler:
    def __init__(
        self,
        potential_energy: Callable[[Q], Union[jnp.ndarray, float]],
        inverse_mass_matrix,
        position_proto: Q,
        step_size: Union[jnp.ndarray, float] = 1.0,
        max_energy_difference: Union[jnp.ndarray, float] = jnp.inf
    ):
        if not callable(potential_energy):
            raise TypeError()
        if not isinstance(step_size, (jnp.ndarray, float)):
            raise TypeError()

        self.potential_energy = potential_energy

        self.inverse_mass_matrix = _parse_diag_mass_matrix(
            inverse_mass_matrix, position_proto=position_proto
        )
        self.mass_matrix_sqrt = self.inverse_mass_matrix**(-0.5)

        self.step_size = step_size

        def kinetic_energy(inverse_mass_matrix, momentum):
            # NOTE, assume a diagonal mass-matrix
            return inverse_mass_matrix.dot(momentum**2) / 2.

        self.kinetic_energy = kinetic_energy
        kinetic_energy_gradient = lambda inv_m, mom: inv_m * mom
        potential_energy_gradient = grad(self.potential_energy)
        self.stepper = partial(
            leapfrog_step, potential_energy_gradient, kinetic_energy_gradient
        )

        self.max_energy_difference = max_energy_difference

        def sample_next_state(key,
                              prev_position: Q) -> Tuple[Any, Tuple[Any, Q]]:
            raise NotImplementedError()

        self.sample_next_state = sample_next_state

    @staticmethod
    def init_chain(
        num_samples: int, position_proto, save_intermediates: bool
    ) -> Chain:
        raise NotImplementedError()

    @staticmethod
    def update_chain(
        chain: Chain, idx: Union[jnp.ndarray, int], tree: Tree
    ) -> Chain:
        raise NotImplementedError()

    def generate_n_samples(
        self,
        key: Any,
        initial_position: Q,
        num_samples,
        *,
        save_intermediates: bool = False
    ) -> Tuple[Chain, Tuple[Any, Q]]:
        if not isinstance(key, (jnp.ndarray, np.ndarray)):
            if isinstance(key, int):
                key = random.PRNGKey(key)
            else:
                raise TypeError()

        chain = self.init_chain(
            num_samples, initial_position, save_intermediates
        )

        def amend_chain(idx, state):
            chain, core_state = state
            tree, core_state = self.sample_next_state(*core_state)
            chain = self.update_chain(chain, idx, tree)
            return chain, core_state

        chain, core_state = fori_loop(
            lower=0,
            upper=num_samples,
            body_fun=amend_chain,
            init_val=(chain, (key, initial_position))
        )

        return chain, core_state


class NUTSChain(_Sampler):
    def __init__(
        self,
        potential_energy: Callable[[Q], Union[float, jnp.ndarray]],
        inverse_mass_matrix,
        position_proto: Q,
        step_size: float = 1.0,
        max_tree_depth: int = 10,
        bias_transition: bool = True,
        max_energy_difference: float = jnp.inf
    ):
        super().__init__(
            potential_energy=potential_energy,
            inverse_mass_matrix=inverse_mass_matrix,
            position_proto=position_proto,
            step_size=step_size,
            max_energy_difference=max_energy_difference
        )

        if not isinstance(max_tree_depth, int):
            raise TypeError()
        self.bias_transition = bias_transition
        self.max_tree_depth = max_tree_depth

        def sample_next_state(key,
                              prev_position: Q) -> Tuple[Tree, Tuple[Any, Q]]:
            key, key_momentum, key_nuts = random.split(key, 3)

            resampled_momentum = sample_momentum_from_diagonal(
                key=key_momentum, mass_matrix_sqrt=self.mass_matrix_sqrt
            )
            qp = QP(position=prev_position, momentum=resampled_momentum)

            tree = generate_nuts_tree(
                initial_qp=qp,
                key=key_nuts,
                step_size=self.step_size,
                max_tree_depth=self.max_tree_depth,
                stepper=self.stepper,
                potential_energy=self.potential_energy,
                kinetic_energy=self.kinetic_energy,
                inverse_mass_matrix=self.inverse_mass_matrix,
                bias_transition=self.bias_transition,
                max_energy_difference=self.max_energy_difference
            )
            return tree, (key, tree.proposal_candidate.position)

        self.sample_next_state = sample_next_state

    @staticmethod
    def init_chain(
        num_samples: int, position_proto, save_intermediates: bool
    ) -> Chain:
        samples = tree_util.tree_map(
            lambda arr: jnp.
            zeros_like(arr, shape=(num_samples, ) + jnp.shape(arr)),
            position_proto
        )
        depths = jnp.zeros(num_samples, dtype=jnp.uint64)
        divergences = jnp.zeros(num_samples, dtype=bool)
        chain = Chain(
            samples=samples,
            divergences=divergences,
            acceptance=0.,
            depths=depths
        )
        if save_intermediates:
            _qp_proto = QP(position_proto, position_proto)
            _tree_proto = Tree(
                _qp_proto,
                _qp_proto,
                0.,
                _qp_proto,
                turning=True,
                diverging=True,
                depth=0,
                cumulative_acceptance=0.
            )
            trees = tree_util.tree_map(
                lambda leaf: jnp.
                zeros_like(leaf, shape=(num_samples, ) + jnp.shape(leaf)),
                _tree_proto
            )
            chain = chain._replace(trees=trees)

        return chain

    @staticmethod
    def update_chain(
        chain: Chain, idx: Union[jnp.ndarray, int], tree: Tree
    ) -> Chain:
        num_proposals = 2**jnp.array(tree.depth, dtype=jnp.uint64) - 1
        tree_acceptance = jnp.where(
            num_proposals > 0, tree.cumulative_acceptance / num_proposals, 0.
        )

        samples = tree_index_update(
            chain.samples, idx, tree.proposal_candidate.position
        )
        divergences = chain.divergences.at[idx].set(tree.diverging)
        depths = chain.depths.at[idx].set(tree.depth)
        acceptance = (
            chain.acceptance + (tree_acceptance - chain.acceptance) / (idx + 1)
        )
        chain = chain._replace(
            samples=samples,
            divergences=divergences,
            acceptance=acceptance,
            depths=depths
        )
        if chain.trees is not None:
            trees = tree_index_update(chain.trees, idx, tree)
            chain = chain._replace(trees=trees)

        return chain


class HMCChain(_Sampler):
    def __init__(
        self,
        potential_energy: Callable,
        inverse_mass_matrix,
        position_proto,
        num_steps,
        step_size: float = 1.0,
        max_energy_difference: float = jnp.inf
    ):
        super().__init__(
            potential_energy=potential_energy,
            inverse_mass_matrix=inverse_mass_matrix,
            position_proto=position_proto,
            step_size=step_size,
            max_energy_difference=max_energy_difference
        )

        if not isinstance(num_steps, (jnp.ndarray, int)):
            raise TypeError()
        self.num_steps = num_steps

        def sample_next_state(key,
                              prev_position: Q) -> Tuple[Tree, Tuple[Any, Q]]:
            key, key_choose, key_momentum_resample = random.split(key, 3)

            resampled_momentum = sample_momentum_from_diagonal(
                key=key_momentum_resample,
                mass_matrix_sqrt=self.mass_matrix_sqrt
            )
            qp = QP(position=prev_position, momentum=resampled_momentum)

            acc_rej = generate_hmc_acc_rej(
                key=key_choose,
                initial_qp=qp,
                potential_energy=self.potential_energy,
                kinetic_energy=self.kinetic_energy,
                inverse_mass_matrix=self.inverse_mass_matrix,
                stepper=self.stepper,
                num_steps=self.num_steps,
                step_size=self.step_size,
                max_energy_difference=self.max_energy_difference
            )
            return acc_rej, (key, acc_rej.accepted_qp.position)

        self.sample_next_state = sample_next_state

    @staticmethod
    def init_chain(
        num_samples: int, position_proto, save_intermediates: bool
    ) -> Chain:
        samples = tree_util.tree_map(
            lambda arr: jnp.
            zeros_like(arr, shape=(num_samples, ) + jnp.shape(arr)),
            position_proto
        )
        divergences = jnp.zeros(num_samples, dtype=bool)
        chain = Chain(samples=samples, divergences=divergences, acceptance=0.)
        if save_intermediates:
            _qp_proto = QP(position_proto, position_proto)
            _acc_rej_proto = AcceptedAndRejected(
                _qp_proto, _qp_proto, True, True
            )
            trees = tree_util.tree_map(
                lambda leaf: jnp.
                zeros_like(leaf, shape=(num_samples, ) + jnp.shape(leaf)),
                _acc_rej_proto
            )
            chain = chain._replace(trees=trees)

        return chain

    @staticmethod
    def update_chain(
        chain: Chain, idx: Union[jnp.ndarray, int], acc_rej: AcceptedAndRejected
    ) -> Chain:
        samples = tree_index_update(
            chain.samples, idx, acc_rej.accepted_qp.position
        )
        divergences = chain.divergences.at[idx].set(acc_rej.diverging)
        acceptance = (
            chain.acceptance + (acc_rej.accepted - chain.acceptance) /
            (idx + 1)
        )
        chain = chain._replace(
            samples=samples, divergences=divergences, acceptance=acceptance
        )
        if chain.trees is not None:
            trees = tree_index_update(chain.trees, idx, acc_rej)
            chain = chain._replace(trees=trees)

        return chain



# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from typing import Callable, Optional, Tuple

import sys
from jax import numpy as jnp
from jax.tree_util import tree_map

from .forest_util import ShapeWithDtype
from .likelihood import Likelihood


def standard_t(nwr, dof):
    return jnp.sum(jnp.log1p(nwr**2 / dof) * (dof + 1)) / 2


def _shape_w_fixed_dtype(dtype):
    def shp_w_dtp(e):
        return ShapeWithDtype(jnp.shape(e), dtype)

    return shp_w_dtp


def _get_cov_inv_and_std_inv(
    cov_inv: Optional[Callable],
    std_inv: Optional[Callable],
    primals=None
) -> Tuple[Callable, Callable]:
    if not cov_inv and not std_inv:

        def cov_inv(tangents):
            return tangents

        def std_inv(tangents):
            return tangents

    elif not cov_inv:
        wm = (
            "assuming a diagonal covariance matrix"
            ";\nsetting `cov_inv` to `std_inv(jnp.ones_like(data))**2`"
        )
        print(wm, file=sys.stderr)
        noise_std_inv_sq = std_inv(tree_map(jnp.ones_like, primals))**2

        def cov_inv(tangents):
            return noise_std_inv_sq * tangents

    elif not std_inv:
        wm = (
            "assuming a diagonal covariance matrix"
            ";\nsetting `std_inv` to `cov_inv(jnp.ones_like(data))**0.5`"
        )
        print(wm, file=sys.stderr)
        noise_cov_inv_sqrt = tree_map(
            jnp.sqrt, cov_inv(tree_map(jnp.ones_like, primals))
        )

        def std_inv(tangents):
            return noise_cov_inv_sqrt * tangents

    if not (callable(cov_inv) and callable(std_inv)):
        raise ValueError("received un-callable input")
    return cov_inv, std_inv


def Gaussian(
    data,
    noise_cov_inv: Optional[Callable] = None,
    noise_std_inv: Optional[Callable] = None
):
    """Gaussian likelihood of the data

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.
    noise_cov_inv : callable acting on type of data
        Function applying the inverse noise covariance of the Gaussian.
    noise_std_inv : callable acting on type of data
        Function applying the square root of the inverse noise covariance.

    Notes
    -----
    If `noise_std_inv` is `None` it is inferred by assuming a diagonal noise
    covariance, i.e. by applying it to a vector of ones and taking the square
    root. If both `noise_cov_inv` and `noise_std_inv` are `None`, a unit
    covariance is assumed.
    """
    noise_cov_inv, noise_std_inv = _get_cov_inv_and_std_inv(
        noise_cov_inv, noise_std_inv, data
    )

    def hamiltonian(primals):
        p_res = primals - data
        return 0.5 * p_res.ravel().dot(noise_cov_inv(p_res).ravel())

    def metric(primals, tangents):
        return noise_cov_inv(tangents)

    def left_sqrt_metric(primals, tangents):
        return noise_std_inv(tangents)

    def transformation(primals):
        return noise_std_inv(primals)

    lsm_tangents_shape = tree_map(ShapeWithDtype.from_leave, data)

    return Likelihood(
        hamiltonian,
        transformation=transformation,
        left_sqrt_metric=left_sqrt_metric,
        metric=metric,
        lsm_tangents_shape=lsm_tangents_shape
    )


def StudentT(
    data,
    dof,
    noise_cov_inv: Optional[Callable] = None,
    noise_std_inv: Optional[Callable] = None
):
    """Student's t likelihood of the data

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.
    dof : tree-like structure of jnp.ndarray and float
        Degree-of-freedom parameter of Student's t distribution.
    noise_cov_inv : callable acting on type of data
        Function applying the inverse noise covariance of the Gaussian.
    noise_std_inv : callable acting on type of data
        Function applying the square root of the inverse noise covariance.

    Notes
    -----
    If `noise_std_inv` is `None` it is inferred by assuming a diagonal noise
    covariance, i.e. by applying it to a vector of ones and taking the square
    root. If both `noise_cov_inv` and `noise_std_inv` are `None`, a unit
    covariance is assumed.
    """
    noise_cov_inv, noise_std_inv = _get_cov_inv_and_std_inv(
        noise_cov_inv, noise_std_inv, data
    )

    def hamiltonian(primals):
        """
        primals : mean
        """
        return standard_t(noise_std_inv(data - primals), dof)

    def metric(primals, tangents):
        """
        primals, tangent : mean
        """
        return noise_cov_inv((dof + 1) / (dof + 3) * tangents)

    def left_sqrt_metric(primals, tangents):
        """
        primals, tangents : mean
        """
        return noise_std_inv(jnp.sqrt((dof + 1) / (dof + 3)) * tangents)

    def transformation(primals):
        """
        primals : mean
        """
        return noise_std_inv(jnp.sqrt((dof + 1) / (dof + 3)) * primals)

    lsm_tangents_shape = tree_map(ShapeWithDtype.from_leave, data)

    return Likelihood(
        hamiltonian,
        transformation=transformation,
        left_sqrt_metric=left_sqrt_metric,
        metric=metric,
        lsm_tangents_shape=lsm_tangents_shape
    )


def Poissonian(data, sampling_dtype=float):
    """Computes the negative log-likelihood, i.e. the Hamiltonians of an
    expected count field constrained by Poissonian count data.

    Represents up to an f-independent term :math:`log(d!)`:

    .. math ::
        E(f) = -\\log \\text{Poisson}(d|f) = \\sum f - d^\\dagger \\log(f),

    where f is a field in data space of the expectation values for the counts.

    Parameters
    ----------
    data : ndarray of uint
        Data field with counts. Needs to have integer dtype and all values need
        to be non-negative.
    sampling_dtype : dtype, optional
        Data-type for sampling.
    """
    from .forest_util import common_type

    dtp = common_type(data)
    if not jnp.issubdtype(dtp, jnp.integer):
        raise TypeError("`data` of invalid type")
    if jnp.any(data < 0):
        raise ValueError("`data` may not be negative")

    def hamiltonian(primals):
        return jnp.sum(primals) - jnp.vdot(jnp.log(primals), data)

    def metric(primals, tangents):
        return tangents / primals

    def left_sqrt_metric(primals, tangents):
        return tangents / jnp.sqrt(primals)

    def transformation(primals):
        return jnp.sqrt(primals) * 2.

    lsm_tangents_shape = tree_map(_shape_w_fixed_dtype(sampling_dtype), data)

    return Likelihood(
        hamiltonian,
        transformation=transformation,
        left_sqrt_metric=left_sqrt_metric,
        metric=metric,
        lsm_tangents_shape=lsm_tangents_shape
    )


def VariableCovarianceGaussian(data):
    """Gaussian likelihood of the data with a variable covariance

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.

    Notes
    -----
    The likelihood acts on a tuple of (mean, std_inv).
    """
    from .sugar import sum_of_squares

    def hamiltonian(primals):
        """
        primals : pair of (mean, std_inv)
        """
        res = (primals[0] - data) * primals[1]
        return 0.5 * sum_of_squares(res) - jnp.sum(jnp.log(primals[1]))

    def metric(primals, tangents):
        """
        primals, tangent : pair of (mean, std_inv)
        """
        prim_std_inv_sq = primals[1]**2
        res = (prim_std_inv_sq * tangents[0], 2 * tangents[1] / prim_std_inv_sq)
        return type(primals)(res)

    def left_sqrt_metric(primals, tangents):
        """
        primals, tangent : pair of (mean, std_inv)
        """
        res = (primals[1] * tangents[0], jnp.sqrt(2) * tangents[1] / primals[1])
        return type(primals)(res)

    def transformation(primals):
        """
        pirmals : pair of (mean, std_inv)

        Notes
        -----
        A global transformation to Euclidean space does not exist. A local
        approximation invoking the residual is used instead.
        """
        # TODO: test by drawing synthetic data that actually follows the
        # noise-cov and then average over it
        res = (primals[1] * (primals[0] - data), tree_map(jnp.log, primals[1]))
        return type(primals)(res)

    lsm_tangents_shape = tree_map(ShapeWithDtype.from_leave, (data, data))

    return Likelihood(
        hamiltonian,
        transformation=transformation,
        left_sqrt_metric=left_sqrt_metric,
        metric=metric,
        lsm_tangents_shape=lsm_tangents_shape
    )


def VariableCovarianceStudentT(data, dof):
    """Student's t likelihood of the data with a variable covariance

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.
    dof : tree-like structure of jnp.ndarray and float
        Degree-of-freedom parameter of Student's t distribution.

    Notes
    -----
    The likelihood acts on a tuple of (mean, std).
    """
    def hamiltonian(primals):
        """
        primals : pair of (mean, std)
        """
        t = standard_t((data - primals[0]) / primals[1], dof)
        t += jnp.sum(jnp.log(primals[1]))
        return t

    def metric(primals, tangent):
        """
        primals, tangent : pair of (mean, std)
        """
        return (
            tangent[0] * (dof + 1) / (dof + 3) / primals[1]**2,
            tangent[1] * 2 * dof / (dof + 3) / primals[1]**2
        )

    def left_sqrt_metric(primals, tangents):
        """
        primals, tangents : pair of (mean, std)
        """
        cov = (
            (dof + 1) / (dof + 3) / primals[1]**2,
            2 * dof / (dof + 3) / primals[1]**2
        )
        res = (jnp.sqrt(cov[0]) * tangents[0], jnp.sqrt(cov[1]) * tangents[1])
        return res

    lsm_tangents_shape = tree_map(ShapeWithDtype.from_leave, (data, data))

    return Likelihood(
        hamiltonian,
        left_sqrt_metric=left_sqrt_metric,
        metric=metric,
        lsm_tangents_shape=lsm_tangents_shape
    )


def Categorical(data, axis=-1, sampling_dtype=float):
    """Categorical likelihood of the data, equivalent to cross entropy

    Parameters
    ----------
    data : sequence of int
        An array stating which of the categories is the realized in the data.
        Must agree with the input shape except for the shape[axis]
    axis : int
        Axis over which the categories are formed
    sampling_dtype : dtype, optional
        Data-type for sampling.
    """
    def hamiltonian(primals):
        from jax.nn import log_softmax
        logits = log_softmax(primals, axis=axis)
        return -jnp.sum(jnp.take_along_axis(logits, data, axis))

    def metric(primals, tangents):
        from jax.nn import softmax

        preds = softmax(primals, axis=axis)
        norm_term = jnp.sum(preds * tangents, axis=axis, keepdims=True)
        return preds * tangents - preds * norm_term

    def left_sqrt_metric(primals, tangents):
        from jax.nn import softmax

        sqrtp = jnp.sqrt(softmax(primals, axis=axis))
        norm_term = jnp.sum(sqrtp * tangents, axis=axis, keepdims=True)
        return sqrtp * (tangents - sqrtp * norm_term)

    lsm_tangents_shape = tree_map(_shape_w_fixed_dtype(sampling_dtype), data)

    return Likelihood(
        hamiltonian,
        left_sqrt_metric=left_sqrt_metric,
        metric=metric,
        lsm_tangents_shape=lsm_tangents_shape
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .. import utilities
from ..domain_tuple import DomainTuple
from ..domains.rg_space import RGSpace
from ..field import Field
from .linear_operator import LinearOperator


class FieldZeroPadder(LinearOperator):
    """Operator which applies zero-padding to one of the subdomains of its
    input field

    Parameters
    ----------
    domain : Domain, DomainTuple or tuple of Domain
        The operator's input domain.
    new_shape : list or tuple of int
        The new dimensions of the subdomain which is zero-padded.
        No entry must be smaller than the corresponding dimension in the
        operator's domain.
    space : int
        The index of the subdomain to be zero-padded. If None, it is set to 0
        if domain contains exactly one space. domain[space] must be an RGSpace.
    central : bool
        If `False`, padding is performed at the end of the domain axes,
        otherwise in the middle.

    Notes
    -----
    When doing central padding on an axis with an even length, the "central"
    entry should in principle be split up; this is currently not done.
    """
    def __init__(self, domain, new_shape, space=0, central=False):
        self._domain = DomainTuple.make(domain)
        self._space = utilities.infer_space(self._domain, space)
        self._central = central
        dom = self._domain[self._space]
        if not isinstance(dom, RGSpace):
            raise TypeError("RGSpace required")
        if len(new_shape) != len(dom.shape):
            raise ValueError("Shape mismatch")
        if any([a < b for a, b in zip(new_shape, dom.shape)]):
            raise ValueError("New shape must not be smaller than old shape")
        self._target = list(self._domain)
        self._target[self._space] = RGSpace(new_shape, dom.distances,
                                            dom.harmonic)
        self._target = DomainTuple.make(self._target)
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        v = x.val
        curshp = list(self._dom(mode).shape)
        tgtshp = self._tgt(mode).shape
        for d in self._target.axes[self._space]:
            if v.shape[d] == tgtshp[d]:  # nothing to do
                continue

            idx = (slice(None),) * d

            if mode == self.TIMES:
                shp = list(v.shape)
                shp[d] = tgtshp[d]
                xnew = np.zeros(shp, dtype=v.dtype)
                if self._central:
                    Nyquist = v.shape[d]//2
                    i1 = idx + (slice(0, Nyquist+1),)
                    xnew[i1] = v[i1]
                    i1 = idx + (slice(None, -(Nyquist+1), -1),)
                    xnew[i1] = v[i1]
#                     if (v.shape[d] & 1) == 0:  # even number of pixels
#                         i1 = idx+(Nyquist,)
#                         xnew[i1] *= 0.5
#                         i1 = idx+(-Nyquist,)
#                         xnew[i1] *= 0.5
                else:
                    xnew[idx + (slice(0, v.shape[d]),)] = v
            else:  # ADJOINT_TIMES
                if self._central:
                    shp = list(v.shape)
                    shp[d] = tgtshp[d]
                    xnew = np.zeros(shp, dtype=v.dtype)
                    Nyquist = xnew.shape[d]//2
                    i1 = idx + (slice(0, Nyquist+1),)
                    xnew[i1] = v[i1]
                    i1 = idx + (slice(None, -(Nyquist+1), -1),)
                    xnew[i1] += v[i1]
#                     if (xnew.shape[d] & 1) == 0:  # even number of pixels
#                         i1 = idx+(Nyquist,)
#                         xnew[i1] *= 0.5
                else:
                    xnew = v[idx + (slice(0, tgtshp[d]),)]

            curshp[d] = xnew.shape[d]
            v = xnew
        return Field(self._tgt(mode), v)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce
import operator
import os
import site
import sys

from setuptools import find_packages, setup

# Workaround until https://github.com/pypa/pip/issues/7953 is fixed
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

exec(open('nifty8/version.py').read())

with open("README.md") as f:
    long_description = f.read()
description = "Library for signal inference algorithms that operate regardless of the underlying grids and their resolutions."

extras_require = {
    "re": ("jax", ),
    "native": ("ducc0>=0.24.0", ),
    "doc": ("sphinx", "pydata-sphinx-theme", "jupyter", "jupytext"),
    "util": ("astropy", ),
}
extras_require["full"] = reduce(operator.add, extras_require.values())

setup(name="nifty8",
      version=__version__,
      author="Martin Reinecke",
      author_email="martin@mpa-garching.mpg.de",
      description=description,
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://ift.pages.mpcdf.de/nifty/",
      project_urls={
          "Bug Tracker": "https://gitlab.mpcdf.mpg.de/ift/nifty/issues",
          "Documentation": "https://ift.pages.mpcdf.de/nifty/",
          "Source Code": "https://gitlab.mpcdf.mpg.de/ift/nifty",
          "Changelog": "https://gitlab.mpcdf.mpg.de/ift/nifty/-/blob/NIFTy_8/ChangeLog",
      },
      packages=find_packages(include=["nifty8", "nifty8.*"]),
      license="GPLv3",
      setup_requires=['scipy>=1.4.1', 'numpy>=1.17'],
      install_requires=['scipy>=1.4.1', 'numpy>=1.17'],
      extras_require=extras_require,
      python_requires='>=3.7',
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Intended Audience :: Science/Research"],
      )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numbers
from functools import reduce
from operator import add
from typing import Callable, Optional
from warnings import warn

import numpy as np

from .. import pointwise
from ..domain_tuple import DomainTuple
from ..logger import logger
from ..multi_domain import MultiDomain
from ..utilities import NiftyMeta, check_object_identity, indent, myassert


class Operator(metaclass=NiftyMeta):
    """Transforms values defined on one domain into values defined on another
    domain, and can also provide the Jacobian.
    """

    @property
    def domain(self):
        """The domain on which the Operator's input Field is defined.

        Returns
        -------
        domain : DomainTuple or MultiDomain
        """
        return self._domain

    @property
    def target(self):
        """The domain on which the Operator's output Field is defined.

        Returns
        -------
        target : DomainTuple or MultiDomain
        """
        return self._target

    @property
    def val(self):
        """The numerical value associated with this object
        For "pure" operators this is `None`. For Field-like objects this
        is a `numpy.ndarray` or a dictionary of `numpy.ndarray`s mathcing the
        object's `target`.

        Returns
        -------
        None or numpy.ndarray or dictionary of np.ndarrays : the numerical value
        """
        return None

    @property
    def jac(self):
        """The Jacobian associated with this object.
        For "pure" operators this is `None`. For Field-like objects this
        can be `None` (in which case the object is a constant), or it can be a
        `LinearOperator` with `domain` and `target` matching the object's.

        Returns
        -------
        None or LinearOperator : the Jacobian

        Notes
        -----
        if `value` is None, this must be `None` as well!
        """
        return None

    @property
    def want_metric(self):
        """Whether a metric should be computed for the full expression.
        This is `False` whenever `jac` is `None`. In other cases it signals
        that operators processing this object should compute the metric.

        Returns
        -------
        bool : whether the metric should be computed
        """
        return False

    @property
    def metric(self):
        """The metric associated with the object.
        This is `None`, except when all the following conditions hold:
        - `want_metric` is `True`

        - `target` is the scalar domain

        - the operator chain contained an operator which could compute the
          metric

        Returns
        -------
        None or LinearOperator : the metric
        """
        return None

    @property
    def jax_expr(self) -> Optional[Callable]:
        """Equivalent representation of the operator in JAX."""
        expr = getattr(self, "_jax_expr", None)
        # NOTE, it is incredibly useful to enable this for debugging
        # if expr is None:
        #     warn(f"no JAX expression associated with operator {self!r}")
        return expr

    def scale(self, factor):
        if not isinstance(factor, numbers.Number):
            raise TypeError(".scale() takes a number as input")
        if factor == 1:
            return self
        from .scaling_operator import ScalingOperator
        return ScalingOperator(self.target, factor)(self)

    def conjugate(self):
        from .simple_linear_operators import ConjugationOperator
        return ConjugationOperator(self.target)(self)

    def sum(self, spaces=None):
        from .contraction_operator import ContractionOperator
        return ContractionOperator(self.target, spaces)(self)

    def integrate(self, spaces=None):
        from .contraction_operator import IntegrationOperator
        return IntegrationOperator(self.target, spaces)(self)

    def vdot(self, other):
        from ..sugar import makeOp
        if not isinstance(other, Operator):
            raise TypeError
        if other.jac is None:
            res = self.conjugate()*other
        else:
            res = makeOp(other) @ self.conjugate()
        return res.sum()

    @property
    def real(self):
        from .simple_linear_operators import Realizer
        return Realizer(self.target)(self)

    @property
    def imag(self):
        from .simple_linear_operators import Imaginizer
        return Imaginizer(self.target)(self)

    def __neg__(self):
        return self.scale(-1)

    def __matmul__(self, x):
        from .energy_operators import LikelihoodEnergyOperator

        if not isinstance(x, Operator):
            return NotImplemented
        if isinstance(x, LikelihoodEnergyOperator):
            return NotImplemented
        return _OpChain.make((self, x))

    def __rmatmul__(self, x):
        from .energy_operators import LikelihoodEnergyOperator

        if not isinstance(x, Operator):
            return NotImplemented
        if isinstance(x, LikelihoodEnergyOperator):
            return NotImplemented
        return _OpChain.make((x, self))

    def partial_insert(self, x):
        from ..multi_domain import MultiDomain
        if not isinstance(x, Operator):
            raise TypeError
        if not isinstance(self.domain, MultiDomain):
            raise TypeError
        if not isinstance(x.target, MultiDomain):
            raise TypeError
        bigdom = MultiDomain.union([self.domain, x.target])
        k1, k2 = set(self.domain.keys()), set(x.target.keys())
        le, ri = k2 - k1, k1 - k2
        leop, riop = self, x
        if len(ri) > 0:
            riop = riop + self.identity_operator(
                MultiDomain.make({kk: bigdom[kk]
                                  for kk in ri}))
        if len(le) > 0:
            leop = leop + self.identity_operator(
                MultiDomain.make({kk: bigdom[kk]
                                  for kk in le}))
        return leop @ riop

    @staticmethod
    def identity_operator(dom):
        from ..sugar import makeDomain
        from .block_diagonal_operator import BlockDiagonalOperator
        from .scaling_operator import ScalingOperator

        dom = makeDomain(dom)
        if isinstance(dom, DomainTuple):
            return ScalingOperator(dom, 1.)
        idops = {kk: ScalingOperator(dd, 1.) for kk, dd in dom.items()}
        return BlockDiagonalOperator(dom, idops)

    def __mul__(self, x):
        if isinstance(x, Operator):
            return _OpProd(self, x)
        if np.isscalar(x):
            return self.scale(x)
        return NotImplemented

    def __rmul__(self, x):
        return self.__mul__(x)

    def __add__(self, x):
        if not isinstance(x, Operator):
            return NotImplemented
        return _OpSum(self, x)

    def __sub__(self, x):
        if not isinstance(x, Operator):
            return NotImplemented
        return _OpSum(self, -x)

    def __abs__(self):
        return self.ptw("abs")

    def __pow__(self, power):
        if not (np.isscalar(power) or power.jac is None):
            return NotImplemented
        return self.ptw("power", power)

    def __getitem__(self, key):
        from ..sugar import is_operator
        from .simple_linear_operators import ducktape

        if not is_operator(self):
            return NotImplemented
        if not isinstance(self.target, MultiDomain):
            raise TypeError("Only Operators with a MultiDomain as target can be subscripted.")
        return ducktape(None, self, key) @ self

    def apply(self, x):
        """Applies the operator to a Field or MultiField.

        Parameters
        ----------
        x : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            Input on which the operator shall act. Needs to be defined on
            :attr:`domain`.
        """
        raise NotImplementedError

    def force(self, x):
        """Extract subset of domain of x according to `self.domain` and apply
        operator."""
        return self.apply(x.extract(self.domain))

    def _check_input(self, x):
        from .scaling_operator import ScalingOperator
        if not (isinstance(x, Operator) and x.val is not None):
            raise TypeError
        if x.jac is not None:
            if not isinstance(x.jac, ScalingOperator):
                raise ValueError
            if x.jac._factor != 1:
                raise ValueError
        check_object_identity(self._domain, x.domain)

    def __call__(self, x):
        if not isinstance(x, Operator):
            raise TypeError
        if x.jac is not None:
            return self.apply(x.trivial_jac()).prepend_jac(x.jac)
        elif x.val is not None:
            return self.apply(x)
        return self @ x

    def ducktape(self, name):
        from ..sugar import is_operator, makeDomain
        from .simple_linear_operators import DomainChangerAndReshaper, ducktape

        if not is_operator(self):
            raise RuntimeError("ducktape works only on operators")

        if isinstance(name, str):  # convert to MultiDomain
            return self @ ducktape(self, None, name)
        else:  # convert domain
            newdom = makeDomain(name)
            return self @ DomainChangerAndReshaper(newdom, self.domain)

    def ducktape_left(self, name):
        from ..sugar import is_fieldlike, is_operator, makeDomain
        from .simple_linear_operators import DomainChangerAndReshaper, ducktape

        if isinstance(name, str):  # convert to MultiDomain
            tgt = self.target if is_operator(self) else self.domain
            return ducktape(None, tgt, name)(self)
        else:  # convert domain
            newdom = DomainTuple.make(name)
            dom = self.domain if is_fieldlike(self) else self.target
            return DomainChangerAndReshaper(dom, newdom)(self)

    def transpose(self, indices):
        """Transposes a Field.

        Parameters
        ----------
        indices : tuple
            Must be a tuple or list which contains a permutation of
            [0,1,..,N-1] where N is the number of domains in the target of the
            Operator (or the Field).
        """
        from ..sugar import is_fieldlike
        from .transpose_operator import TransposeOperator

        dom = self.domain if is_fieldlike(self) else self.target
        return TransposeOperator(dom, indices)(self)

    def __repr__(self):
        return self.__class__.__name__

    def simplify_for_constant_input(self, c_inp):
        from ..multi_field import MultiField
        from ..sugar import makeDomain
        from .energy_operators import EnergyOperator, LikelihoodEnergyOperator
        from .simplify_for_const import (ConstantEnergyOperator,
                                         ConstantLikelihoodEnergyOperator,
                                         ConstantOperator)
        if c_inp is None or (isinstance(c_inp, MultiField) and len(c_inp.keys()) == 0):
            return None, self
        dom = c_inp.domain
        if isinstance(dom, MultiDomain) and len(dom) == 0:
            return None, self

        # Convention: If c_inp is MultiField, it needs to be defined on a
        # subdomain of self._domain
        if isinstance(self.domain, MultiDomain):
            myassert(isinstance(dom, MultiDomain))
            if not set(c_inp.keys()) <= set(self.domain.keys()):
                raise ValueError

        if dom is self.domain:
            if isinstance(self, DomainTuple):
                raise RuntimeError
            if isinstance(self, EnergyOperator):
                op = ConstantEnergyOperator(self(c_inp))
            elif isinstance(self, LikelihoodEnergyOperator):
                op = ConstantLikelihoodEnergyOperator(self(c_inp))
            else:
                op = ConstantOperator(self(c_inp))
            return None, op
        if not isinstance(dom, MultiDomain):
            raise RuntimeError
        c_out, op = self._simplify_for_constant_input_nontrivial(c_inp)
        vardom = makeDomain({kk: vv for kk, vv in self.domain.items()
                             if kk not in c_inp.keys()})
        myassert(op.domain is vardom)
        myassert(op.target is self.target)
        myassert(isinstance(op, Operator))
        if c_out is not None:
            myassert(isinstance(c_out, MultiField))
            myassert(len(set(c_out.keys()) & self.domain.keys()) == 0)
            myassert(set(c_out.keys()) <= set(c_inp.keys()))
        return c_out, op

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from .simplify_for_const import InsertionOperator
        logger.warning('SlowPartialConstantOperator used for:')
        logger.warning(self.__repr__())
        return None, self @ InsertionOperator(self.domain, c_inp)

    def ptw(self, op, *args, **kwargs):
        return _OpChain.make((_FunctionApplier(self.target, op, *args, **kwargs), self))

    def ptw_pre(self, op, *args, **kwargs):
        return _OpChain.make((self, _FunctionApplier(self.domain, op, *args, **kwargs)))

    def apply_to_random_sample(self, **kwargs):
        """Applies the operator to a sample drawn with `ift.sugar.from_random`.
        Keyword arguments are passed through to the sample generation.

        .. warning::
            If not specified otherwise, the sample will be of dtype
            `np.float64`. If the operator requires input values of other
            dtypes, this needs to be indicated with the `dtype` keyword argument.
        """
        from ..sugar import from_random
        random_input = from_random(self.domain, **kwargs)
        return self(random_input)


for f in pointwise.ptw_dict.keys():
    def func(f):
        def func2(self, *args, **kwargs):
            return self.ptw(f, *args, **kwargs)
        return func2
    setattr(Operator, f, func(f))
    def func(f):
        def func2(self, *args, **kwargs):
            return self.ptw_pre(f, *args, **kwargs)
        return func2
    setattr(Operator, f + "_pre", func(f))


class _FunctionApplier(Operator):
    def __init__(self, domain, funcname, *args, **kwargs):
        from ..sugar import makeDomain
        self._domain = self._target = makeDomain(domain)
        self._funcname = funcname
        self._args = args
        self._kwargs = kwargs

        try:
            import jax.numpy as jnp
            from jax import nn as jax_nn

            if funcname in pointwise.ptw_nifty2jax_dict:
                jax_expr = pointwise.ptw_nifty2jax_dict[funcname]
            elif hasattr(jnp, funcname):
                jax_expr = getattr(jnp, funcname)
            elif hasattr(jax_nn, funcname):
                jax_expr = getattr(jax_nn, funcname)
            else:
                warn(f"unable to add JAX call for {funcname!r}")
                jax_expr = None

            def jax_expr_part(x):  # Partial insert with first open argument
                return jax_expr(x, *args, **kwargs)

            if isinstance(self.domain, MultiDomain):
                from functools import partial

                from jax.tree_util import tree_map

                jax_expr_part = partial(tree_map, jax_expr_part)
            self._jax_expr = jax_expr_part
        except ImportError:
            self._jax_expr = None

    def apply(self, x):
        self._check_input(x)
        return x.ptw(self._funcname, *self._args, **self._kwargs)

    def __repr__(self):
        return f"_FunctionApplier ('{self._funcname}')"


class _CombinedOperator(Operator):
    def __init__(self, ops, jax_ops, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError
        self._ops = tuple(ops)

        if all(callable(jop) for jop in jax_ops):

            def joined_jax_op(x):
                for jop in reversed(jax_ops):
                    x = jop(x)
                return x

            self._jax_expr = joined_jax_op
        else:
            self._jax_expr = None

    @classmethod
    def unpack(cls, ops, res):
        for op in ops:
            if isinstance(op, cls):
                res = cls.unpack(op._ops, res)
            else:
                res = res + [op]
        return res

    @classmethod
    def make(cls, ops):
        res = cls.unpack(ops, [])
        if len(res) == 1:
            return res[0]
        jax_res = tuple(op.jax_expr for op in ops)
        return cls(res, jax_res, _callingfrommake=True)


class _OpChain(_CombinedOperator):
    def __init__(self, ops, jax_ops, _callingfrommake=False):
        super(_OpChain, self).__init__(ops, jax_ops, _callingfrommake)
        self._domain = self._ops[-1].domain
        self._target = self._ops[0].target
        for i in range(1, len(self._ops)):
            check_object_identity(self._ops[i-1].domain, self._ops[i].target)

    def apply(self, x):
        self._check_input(x)
        for op in reversed(self._ops):
            x = op(x)
        return x

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from ..multi_domain import MultiDomain
        if not isinstance(self._domain, MultiDomain):
            return None, self
        newop = None
        for op in reversed(self._ops):
            c_inp, t_op = op.simplify_for_constant_input(c_inp)
            newop = t_op if newop is None else op(newop)
        return c_inp, newop

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in self._ops)
        return "_OpChain:\n" + indent(subs)


class _OpProd(Operator):
    def __init__(self, op1, op2):
        from ..sugar import domain_union
        self._domain = domain_union((op1.domain, op2.domain))
        self._target = op1.target
        if op1.target != op2.target:
            raise ValueError("target mismatch")
        self._op1 = op1
        self._op2 = op2

        lhs_has_jax = callable(self._op1.jax_expr)
        rhs_has_jax = callable(self._op2.jax_expr)
        if lhs_has_jax and rhs_has_jax:

            def joined_jax_expr(x):
                return self._op1.jax_expr(x) * self._op2.jax_expr(x)

            self._jax_expr = joined_jax_expr
        else:
            self._jax_expr = None

    def apply(self, x):
        from ..linearization import Linearization
        from ..sugar import makeOp
        self._check_input(x)
        lin = x.jac is not None
        wm = x.want_metric if lin else False
        x = x.val if lin else x
        v1 = x.extract(self._op1.domain)
        v2 = x.extract(self._op2.domain)
        if not lin:
            return self._op1(v1) * self._op2(v2)
        lin1 = self._op1(Linearization.make_var(v1, wm))
        lin2 = self._op2(Linearization.make_var(v2, wm))
        jac = (makeOp(lin1._val)(lin2._jac))._myadd(makeOp(lin2._val)(lin1._jac), False)
        return lin1.new(lin1._val*lin2._val, jac)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from ..multi_domain import MultiDomain
        from .simplify_for_const import ConstCollector
        f1, o1 = self._op1.simplify_for_constant_input(
            c_inp.extract_part(self._op1.domain))
        f2, o2 = self._op2.simplify_for_constant_input(
            c_inp.extract_part(self._op2.domain))
        if not isinstance(self._target, MultiDomain):
            return None, _OpProd(o1, o2)
        cc = ConstCollector()
        cc.mult(f1, o1.target)
        cc.mult(f2, o2.target)
        return cc.constfield, _OpProd(o1, o2)

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in (self._op1, self._op2))
        return "_OpProd:\n"+indent(subs)


class _OpSum(Operator):
    def __init__(self, op1, op2):
        from ..sugar import domain_union
        self._domain = domain_union((op1.domain, op2.domain))
        self._target = domain_union((op1.target, op2.target))
        self._op1 = op1
        self._op2 = op2

        try:
            from ..re import unite

            def joined_jax_expr(x):
                return unite(self._op1.jax_expr(x), self._op2.jax_expr(x))

            self._jax_expr = joined_jax_expr
        except ImportError:
            self._jax_expr = None

    def apply(self, x):
        self._check_input(x)
        return self._apply_operator_sum(x, [self._op1, self._op2])

    @staticmethod
    def _apply_operator_sum(x, ops):
        from ..linearization import Linearization

        unite = lambda x, y: x.unite(y)
        if x.jac is None:
            return reduce(unite, (oo.force(x) for oo in ops))
        lin = [oo(Linearization.make_var(x.val.extract(oo.domain), x.want_metric))
                for oo in ops]
        jacs = map(lambda x: x._jac, lin)
        vals = map(lambda x: x._val, lin)
        metrics = list(map(lambda x: x._metric, lin))
        jac = reduce(lambda x, y: x._myadd(y, False), jacs)
        val = reduce(unite, vals)
        res = x.new(val, jac)
        if all(mm is not None for mm in metrics):
            res = res.add_metric(reduce(add, metrics))
        return res

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from ..multi_domain import MultiDomain
        from .simplify_for_const import ConstCollector
        f1, o1 = self._op1.simplify_for_constant_input(
            c_inp.extract_part(self._op1.domain))
        f2, o2 = self._op2.simplify_for_constant_input(
            c_inp.extract_part(self._op2.domain))
        if not isinstance(self._target, MultiDomain):
            return None, _OpSum(o1, o2)
        cc = ConstCollector()
        cc.add(f1, o1.target)
        cc.add(f2, o2.target)
        return cc.constfield, _OpSum(o1, o2)

    def __repr__(self):
        subs = "\n".join(sub.__repr__() for sub in (self._op1, self._op2))
        return "_OpSum:\n"+indent(subs)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Copyright(C) 2022 Max-Planck-Society, Philipp Arras
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import pickle
from functools import reduce
from os import makedirs
from os.path import isdir, isfile, join
from warnings import warn

import numpy as np

from ..domain_tuple import DomainTuple
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from ..operators.counting_operator import CountingOperator
from ..operators.energy_operators import StandardHamiltonian
from ..operators.operator import Operator
from ..operators.scaling_operator import ScalingOperator
from ..plot import Plot, plottable2D
from ..sugar import from_random
from ..utilities import (Nop, check_MPI_equality,
                         check_MPI_synced_random_state, check_object_identity,
                         get_MPI_params_from_comm)
from .energy_adapter import EnergyAdapter
from .iteration_controllers import EnergyHistory, IterationController
from .kl_energies import SampledKLEnergy
from .minimizer import Minimizer
from .sample_list import (ResidualSampleList, SampleList, SampleListBase,
                          _barrier)

try:
    import h5py
except ImportError:
    h5py = False

try:
    import astropy
except ImportError:
    astropy = False

_output_directory = None
_save_strategy = None


def optimize_kl(likelihood_energy,
                total_iterations,
                n_samples,
                kl_minimizer,
                sampling_iteration_controller,
                nonlinear_sampling_minimizer,
                constants=[],
                point_estimates=[],
                transitions=None,
                export_operator_outputs={},
                output_directory=None,
                initial_position=None,
                initial_index=0,
                comm=None,
                inspect_callback=None,
                terminate_callback=None,
                plot_energy_history=True,
                plot_minisanity_history=True,
                save_strategy="last",
                return_final_position=False,
                resume=False,
                sanity_checks=True):
    """Provide potentially useful interface for standard KL minimization.

    The parameters `likelihood_energy`, `kl_minimizer`,
    `sampling_iteration_controller`, `nonlinear_sampling_minimizer`, `constants`,
    `point_estimates` and `comm` also accept a function as input that takes the
    index of the global iteration as input and returns the respective value that
    should be used for that iteration.

    High-level pseudo code of the algorithm, that is implemented by `optimize_kl`:

    .. code-block:: none

        for ii in range(initial_index, total_iterations):
            samples = Draw `n_samples` approximate samples(position,
                                                           likelihood_energy
                                                           sampling_iteration_controller,
                                                           nonlinear_sampling_minimizer,
                                                           point_estimates)
            position, samples = Optimize approximate KL(likelihood_energy, samples) with `kl_minimizer`(constants)
            Save intermediate results(samples)

    Parameters
    ----------
    likelihood_energy : Operator or callable
        Likelihood energy shall be used for inference. It is assumed that the
        definition space of this energy contains parameters that are a-priori
        standard normal distributed. It needs to have a MultiDomain as Domain.
    total_iterations : int
        Number of resampling loops.
    n_samples : int or callable
        Number of samples used to sample Kullback-Leibler divergence. 0
        corresponds to maximum-a-posteriori.
    kl_minimizer : Minimizer or callable
        Controls the minimizer for the KL optimization.
    sampling_iteration_controller : IterationController or None or callable
        Controls the conjugate gradient for inverting the posterior metric.  If
        `None`, approximate posterior samples cannot be drawn. It is only
        suited for maximum-a-posteriori solutions.
    nonlinear_sampling_minimizer : Minimizer or None or callable
        Controls the minimizer for the non-linear geometric sampling to
        approximate the KL. Can be either None (then the MGVI algorithm is used
        instead of geoVI) or a Minimizer.
    constants : list or callable
        List of parameter keys that are kept constant during optimization.
        Default is no constants.
    point_estimates : list or callable
        List of parameter keys for which no samples are drawn, but that are
        (possibly) optimized for, corresponding to point estimates of these.
        Default is to draw samples for the complete domain.
    transitions : Operator or callable
        Operator that is applied at the beginning of each iteration to the
        latent position. This can be used if parts of the likelihood_energy are
        replaced and thereby have a different domain. If None is passed, no
        transition will be applied. Default is None.
    export_operator_outputs : dict
        Dictionary of operators that are exported during the minimization. The
        key contains a string that serves as identifier. The value of the
        dictionary is an operator.
    output_directory : str or None
        Directory in which all output files are saved. If None, no output is
        stored.  Default: None.
    initial_position : :class:`nifty8.field.Field`, :class:`nifty8.multi_field.MultiField` or None
        Position in the definition space of `likelihood_energy` from which the
        optimization is started. If `None`, it starts at a random, normal
        distributed position with standard deviation 0.1. Default: None.
    initial_index : int
        Initial index that is used to enumerate the output files. May be used
        if `optimize_kl` is called multiple times. Default: 0.
    comm : MPI communicator or None
        MPI communicator for distributing samples over MPI tasks. If `None`,
        the samples are not distributed. Default: None.
    inspect_callback : callable or None
        Function that is called after every global iteration. It can be either a
        function with one argument (then the latest sample list is passed), a
        function with two arguments (in which case the latest sample list and
        the global iteration index are passed). Default: None.
    terminate_callback : callable or None
        Function that is called after every global iteration and after
        `inspect_callback` if present.  It can be either None or a function
        that takes the global iteration index as input and returns a boolean.
        If the return value is true, the global loop in `optimize_kl` is
        terminated. Default: None.
    plot_energy_history : bool
        Determine if the KLEnergy values shall be plotted or not. Default: True.
    plot_minisanity_history : bool
        Determine if the reduced chi-square values computed by minisanity shall
        be plotted or not. Default: True.
    save_strategy : str
        If "last", only the samples of the last global iteration are stored. If
        "all", all intermediate samples are written to disk. `save_strategy` is
        only applicable if `output_directory` is not None. Default: "last".
    return_final_position : bool
        Determine if the final position of the minimization shall be return.
        May be useful to feed it as `initial_position` into another
        `optimize_kl` call. Default: False.
    resume : bool
        Resume partially run optimization. If `True` and `output_directory`
        contains `last_finished_iteration`, `initial_index` and
        `initial_position` are ignored and read from the output directory
        instead. If `last_finished_iteration` is not a file, the value of
        `initial_position` is used instead. Default: False.
    sanity_checks : bool
        Check if all domains of likelihoods and transitions fit together. This
        is potentially expensive because all likelihoods have to be instantiated
        multiple times. Default: True.

    Returns
    -------
    sl : SampleList

    mean : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField` (optional)

    Note
    ----
    If `h5py` is available, the output of all plotted operators is saved as hdf5
    file as well.  If `astropy` is available, the mean and standard deviation of
    all plotted operators, that have a single 2d-RGSpace as target, are exported
    as FITS files.

    Note
    ----
    This function comes with some MPI support. Generally, with the help of MPI
    samples are distributed over tasks.
    """
    from ..utilities import myassert
    from .descent_minimizers import DescentMinimizer

    if not isinstance(export_operator_outputs, dict):
        raise TypeError
    if len(set(["pickle"]) & set(export_operator_outputs.keys())) != 0:
        raise ValueError("The key `pickle` in `export_operator_outputs` is reserved.")
    if not isinstance(initial_index, int):
        raise TypeError
    if save_strategy not in ["all", "last"]:
        raise ValueError(f"Save strategy '{save_strategy}' not supported.")
    if output_directory is None and resume:
        raise ValueError("Can only resume minimization if output_directory is not None")

    likelihood_energy = _make_callable(likelihood_energy)
    kl_minimizer = _make_callable(kl_minimizer)
    sampling_iteration_controller = _make_callable(sampling_iteration_controller)
    nonlinear_sampling_minimizer = _make_callable(nonlinear_sampling_minimizer)
    constants = _make_callable(constants)
    point_estimates = _make_callable(point_estimates)
    transitions = _make_callable(transitions)
    n_samples = _make_callable(n_samples)
    comm = _make_callable(comm)
    if inspect_callback is None:
        inspect_callback = lambda x: None
    if terminate_callback is None:
        terminate_callback = lambda x: False

    if output_directory is not None:
        global _output_directory
        global _save_strategy
        _output_directory = output_directory
        _save_strategy = save_strategy

        lfile = join(output_directory, "last_finished_iteration")
        if resume and isfile(lfile):
            with open(lfile) as f:
                last_finished_index = int(f.read())
            initial_index = last_finished_index + 1
            fname = _file_name_by_strategy(last_finished_index)
            fname = reduce(join, [output_directory, "pickle", fname])
            if isfile(fname + ".mean.pickle"):
                initial_position = ResidualSampleList.load_mean(fname)
            else:
                sl = SampleList.load(fname)
                myassert(sl.n_samples == 1)
                initial_position = sl.local_item(0)
            _load_random_state(last_finished_index)
            energy_history = _pickle_load_values(last_finished_index, 'energy_history')

            if initial_index == total_iterations:
                if isfile(fname + ".mean.pickle"):
                    sl = ResidualSampleList.load(fname)
                return (sl, initial_position) if return_final_position else sl

    if sanity_checks:
        if initial_index >= total_iterations:
            raise ValueError("Initial index is bigger than total iterations: "
                            f"{initial_index} >= {total_iterations}")

        for iglobal in range(initial_index, total_iterations):
            for (obj, cls) in [(likelihood_energy, Operator), (kl_minimizer, DescentMinimizer),
                            (nonlinear_sampling_minimizer, (DescentMinimizer, type(None))),
                            (constants, (list, tuple)), (point_estimates, (list, tuple)),
                            (transitions, (Operator, type(None))), (n_samples, int)]:
                if not isinstance(obj(iglobal), cls):
                    raise TypeError(f"{obj(iglobal)} is not instance of {cls} but rather {type(obj(iglobal))}")

            if sampling_iteration_controller(iglobal) is None:
                myassert(n_samples(iglobal) == 0)
            else:
                myassert(isinstance(sampling_iteration_controller(iglobal), IterationController))
            myassert(likelihood_energy(iglobal).target is DomainTuple.scalar_domain())
            if not comm(iglobal) is None:
                try:
                    import mpi4py
                    myassert(isinstance(comm(iglobal), mpi4py.MPI.Intracomm))
                except ImportError:
                    pass
        myassert(_number_of_arguments(inspect_callback) in [1, 2])
        myassert(_number_of_arguments(terminate_callback) == 1)

        if not likelihood_energy(initial_index).target is DomainTuple.scalar_domain():
            raise TypeError

    # Initial position
    check_MPI_synced_random_state(comm(initial_index))
    if initial_position is None:
        from ..sugar import full, makeDomain
        mean = full(makeDomain({}), 0.)
    else:
        mean = initial_position
    # /Initial position

    # Automatic transitions
    def trans(iglobal, prev_dom, next_dom):
        res = transitions(iglobal)
        if res is not None:
            return res

        if iglobal == 0:
            dom = mean.domain
        else:
            dom = prev_dom
        tgt = next_dom

        if dom is tgt:
            return Operator.identity_operator(dom)
        if set(dom.keys()).issubset(set(tgt.keys())):
            return _InitializeOperator(dom, tgt, std=0.1)
        raise RuntimeError

    if sanity_checks:
        for iglobal in range(initial_index, total_iterations):
            next_dom = likelihood_energy(iglobal).domain
            if iglobal == 0:
                prev_dom = mean.domain
            else:
                prev_dom = likelihood_energy(iglobal-1).domain

            next_dom1 = trans(iglobal, prev_dom, next_dom).target
            if next_dom != next_dom1:
                raise RuntimeError(f"The domain of lh energy #{iglobal} should equal the target of transition.\n"
                                f"{next_dom}\n\n"
                                f"{next_dom1}")
            myassert(prev_dom is trans(iglobal, prev_dom, next_dom).domain)
    # /Automatic transitions

    if output_directory is not None:
        # Create all necessary subfolders
        if _MPI_master(comm(initial_index)):
            makedirs(output_directory, exist_ok=True)
            subfolders = ["pickle"] + list(export_operator_outputs.keys())
            if plot_energy_history:
                subfolders += ["energy_history"]
            if plot_minisanity_history:
                subfolders += ["minisanity_history"]
            for subfolder in subfolders:
                makedirs(join(output_directory, subfolder), exist_ok=True)

    if initial_index == 0:
        energy_history = EnergyHistory()

    for iglobal in range(initial_index, total_iterations):
        lh = likelihood_energy(iglobal)
        mean = trans(iglobal, mean.domain, lh.domain)(mean)
        dom = mean.domain
        mf_dom = isinstance(dom, MultiDomain)
        if not mf_dom:
            warn("In the upcoming release nifty8.optimize_kl will no longer support "
                 "DomainTuple as latent position. You can use "
                 "`likelihood_energy.ducktape(key)` to convert a DomainTuple domain into "
                 "a MultiDomain. Please report at `c@philipp-arras.de` if you use this "
                 "feature and would like to see it continued.", DeprecationWarning)

        count = CountingOperator(lh.domain)
        ham = StandardHamiltonian(lh @ count, sampling_iteration_controller(iglobal))
        minimizer = kl_minimizer(iglobal)
        mean_iter = mean.extract(ham.domain)

        # TODO Distributing the domain of the likelihood is not supported (yet)
        check_MPI_synced_random_state(comm(iglobal))
        check_MPI_equality(lh.domain, comm(iglobal))
        check_MPI_equality(mean.domain, comm(iglobal))
        check_MPI_equality(mean, comm(iglobal))

        if n_samples(iglobal) == 0:
            e = EnergyAdapter(mean_iter, ham, constants=constants(iglobal),
                              want_metric=_want_metric(minimizer))
            if comm(iglobal) is None:
                e, _ = minimizer(e)
                mean = MultiField.union([mean, e.position]) if mf_dom else e.position
                sl = SampleList([mean])
                energy_history.append((iglobal, e.value))
            else:
                warn("Have detected MPI communicator for optimizing Hamiltonian. Will use only "
                     "the rank0 task and communicate the result afterwards to all other tasks.")
                if _MPI_master(comm(iglobal)):
                    e, _ = minimizer(e)
                    mean = MultiField.union([mean, e.position]) if mf_dom else e.position
                    sl = SampleList([mean], comm=comm(iglobal), domain=dom)
                    energy_history.append((iglobal, e.value))
                else:
                    mean = None
                    sl = SampleList([], comm=comm(iglobal), domain=dom)
                _barrier(comm(iglobal))
                mean = comm(iglobal).bcast(mean, root=0)
        else:
            e = SampledKLEnergy(
                mean_iter,
                ham,
                n_samples(iglobal),
                nonlinear_sampling_minimizer(iglobal),
                comm=comm(iglobal),
                constants=constants(iglobal),
                point_estimates=point_estimates(iglobal))
            e, _ = minimizer(e)
            mean = MultiField.union([mean, e.position]) if mf_dom else e.position
            sl = e.samples.at(mean)
            energy_history.append((iglobal, e.value))

        _minisanity(lh, iglobal, sl, comm, plot_minisanity_history)
        _barrier(comm(iglobal))

        if output_directory is not None:
            _plot_operators(iglobal, export_operator_outputs, sl, comm(iglobal))
            sl.save(join(output_directory, "pickle/") + _file_name_by_strategy(iglobal),
                    overwrite=True)
            _save_random_state(iglobal)

            if _MPI_master(comm(iglobal)):
                with open(join(output_directory, "last_finished_iteration"), "w") as f:
                    f.write(str(iglobal))
                _pickle_save_values(iglobal, 'energy_history', energy_history)
                if plot_energy_history:
                    _plot_energy_history(iglobal, energy_history)
        _barrier(comm(iglobal))

        _counting_report(count, iglobal, comm)

        _handle_inspect_callback(inspect_callback, sl, iglobal)
        _barrier(comm(iglobal))

        if _handle_terminate_callback(terminate_callback, iglobal, comm):
            break
        _barrier(comm(iglobal))

        del lh

    return (sl, mean) if return_final_position else sl


def _file_name(name, index, prefix=""):
    op_direc = join(_output_directory, name)
    return join(op_direc, f"{prefix}{index:03d}.png")


def _file_name_by_strategy(iglobal, save_strategy='global_strategy'):
    if save_strategy == 'global_strategy':
        save_strategy = _save_strategy
    if save_strategy == "all":
        return f"iteration_{iglobal}"
    elif save_strategy == "last":
        return "last"
    raise RuntimeError


def _save_random_state(index):
    from ..random import getState
    file_name = join(_output_directory, "pickle/nifty_random_state_")
    file_name += _file_name_by_strategy(index)
    with open(file_name, "wb") as f:
        f.write(getState())


def _load_random_state(index):
    from ..random import setState
    file_name = join(_output_directory, "pickle/nifty_random_state_")
    file_name += _file_name_by_strategy(index)
    with open(file_name, "rb") as f:
        setState(f.read())


def _pickle_save_values(index, name, val):
    file_name = join(_output_directory, f"pickle/{name}_")
    file_name += _file_name_by_strategy(index)
    with open(file_name, "wb") as f:
        pickle.dump(val, f)


def _pickle_load_values(index, name):
    file_name = join(_output_directory, f"pickle/{name}_")
    file_name += _file_name_by_strategy(index)
    with open(file_name, "rb") as f:
        val = pickle.load(f)
    return val


def _plot_operators(index, export_operator_outputs, sample_list, comm):
    if not isinstance(export_operator_outputs, dict):
        raise TypeError
    if not isdir(_output_directory):
        raise RuntimeError(f"{_output_directory} does not exist")
    if not isinstance(sample_list, SampleListBase):
        raise TypeError

    for name, op in export_operator_outputs.items():
        if not _is_subdomain(op.domain, sample_list.domain):
            continue

        op_direc = join(_output_directory, name)
        if sample_list.n_samples > 1:
            cfg = {"samples": True, "mean": True, "std": True}
        else:
            cfg = {"samples": True, "mean": False, "std": False}

        if h5py:
            file_name = join(op_direc, _file_name_by_strategy(index) + ".hdf5")
            sample_list.save_to_hdf5(file_name, op=op, overwrite=True, **cfg)

        if astropy:
            try:
                file_name_base = join(op_direc, _file_name_by_strategy(index))
                sample_list.save_to_fits(file_name_base, op=op, overwrite=True, **cfg)
            except ValueError:
                pass


def _plot_energy_history(index, energy_history):
    import matplotlib.pyplot as plt
    fname = join(_output_directory, 'energy_history',
                 '{}_' + _file_name_by_strategy(index) + '.png')

    E = np.array(energy_history.energy_values)

    # energy value plot
    p = Plot()
    p.add(energy_history, skip_timestamp_conversion=True, xlabel='iteration',
          ylabel=r'E', yscale='log' if (E > 0.).all() else 'linear')
    p.output(title='energy history', name=fname.format('energy_history'))

    # energy change plot
    if index > 0:
        ts = np.array(energy_history.time_stamps[1:])
        dE = E[1:] - E[:-1]
        idx_pos = (dE > 0)
        idx_neg = (dE < 0)
        plt.plot(ts[idx_pos], dE[idx_pos], '^', color='red', label='positive')
        plt.plot(ts[idx_neg], -dE[idx_neg], 'v', color='green', label='negative')
        plt.yscale('log')
        plt.title('energy change w.r.t. previous step')
        plt.xlabel('iteration')
        plt.ylabel(r'$\Delta\,$E')
        plt.legend()
        plt.savefig(fname.format('energy_change_history'))
        plt.clf()


def _append_key(s, key):
    if key == "":
        return s
    return f"{s} ({key})"


def _minisanity(likelihood_energy, iglobal, sl, comm, plot_minisanity_history):
    from ..extra import minisanity

    s, ms_val = minisanity(likelihood_energy, sl, terminal_colors=False,
                           return_values=True)
    check_MPI_equality(ms_val, comm(iglobal))
    _report_to_logger_and_file(s, "minisanity.txt", iglobal, comm, True, True,
                               True)

    if _MPI_master(comm(iglobal)) and _output_directory is not None:
        # load/create minisanity_history object
        value_type_keys = ['redchisq', 'scmean']
        category_keys = ['data_residuals', 'latent_variables']

        if iglobal == 0:
            mh = {tk: {ck: {} for ck in category_keys} for tk in value_type_keys}
        else:
            mh = _pickle_load_values(iglobal - 1, 'minisanity_history')

        key_set_mh = {}
        key_set_msval = {}
        for ck in category_keys:
            key_set_mh[ck] = set(mh['redchisq'][ck].keys())
            key_set_msval[ck] = set(ms_val['redchisq'][ck].keys())

        for tk in value_type_keys:
            for ck in category_keys:
                # all keys not yet in minisanity history
                for ek in key_set_msval[ck] - key_set_mh[ck]:
                    v = ms_val[tk][ck][ek]
                    mh[tk][ck][ek] = {}
                    mh[tk][ck][ek]['index'] = [iglobal, ]
                    mh[tk][ck][ek]['mean'] = [v['mean'], ]
                    mh[tk][ck][ek]['std'] = [v['std'], ]
                # all keys already present in minisanity history
                for ek in key_set_msval[ck] & key_set_mh[ck]:
                    v = ms_val[tk][ck][ek]
                    mh[tk][ck][ek]['index'].append(iglobal)
                    mh[tk][ck][ek]['mean'].append(v['mean'])
                    mh[tk][ck][ek]['std'].append(v['std'])

        _pickle_save_values(iglobal, 'minisanity_history', mh)

        if plot_minisanity_history:
            _plot_minisanity_history(iglobal, mh)


def _plot_minisanity_history(index, minisanity_history):
    import matplotlib.pyplot as plt
    from matplotlib.cm import plasma

    mhrcs = minisanity_history['redchisq']

    labels = []
    vals = []
    idxs = []

    n_dr = 0
    for kk in mhrcs['data_residuals'].keys():
        labels.append(f'residuals: {n_dr}')
        vals.append(mhrcs['data_residuals'][kk]['mean'])
        idxs.append(mhrcs['data_residuals'][kk]['index'])
        n_dr += 1

    n_lv = 0
    for kk in mhrcs['latent_variables'].keys():
        labels.append(f'latent: {kk}')
        vals.append(mhrcs['latent_variables'][kk]['mean'])
        idxs.append(mhrcs['latent_variables'][kk]['index'])
        n_lv += 1

    n_tot = n_dr + n_lv

    colors = [plasma(x) for x in np.linspace(0, 0.95, n_tot)]

    linestyles = ['-'] * n_tot
    for ii in range(1, n_tot, 2):
        linestyles[ii] = '--'

    vals = [np.array(v) for v in vals]
    idxs = [np.array(i) for i in idxs]

    plt.figure()
    for i in range(n_tot):
        plt.plot(idxs[i], vals[i], label=labels[i], color=colors[i], marker='.',
                 linestyle=linestyles[i])

    xlim = plt.xlim()
    plt.hlines(1., xlim[0], xlim[1], linestyle='dashed', color='black', linewidth=2, zorder=-1)
    plt.xlim(*xlim)
    plt.yscale('log')

    plt.title(r'reduced $\chi^2$ values')
    plt.xlabel('iteration')
    plt.ylabel(r'red. $\chi^2$')
    plt.legend(loc='upper right')
    plt.savefig(join(_output_directory, 'minisanity_history',
                     'minisanity_history_' + _file_name_by_strategy(index) + '.png'))
    plt.clf()


def _counting_report(count, iglobal, comm):
    _report_to_logger_and_file(count.report(), "counting_report.txt", iglobal,
                               comm, _output_directory is None, True, False)


def _report_to_logger_and_file(report, file_name, iglobal, comm, to_logger,
                               to_file, only_master):
    from datetime import datetime

    from ..logger import logger
    from ..utilities import allreduce_sum

    intro = f"Finished index: {iglobal}\nCurrent datetime: {datetime.now()}\n"

    if not only_master:
        report = allreduce_sum([[report]], comm(iglobal))
        report = [f"Task {ii}\n{rr}" for ii, rr in enumerate(report)]
        report = "\n".join(report)

    if _MPI_master(comm(iglobal)):
        if to_logger:
            logger.info(report)
        if _output_directory is not None and to_file:
            with open(join(_output_directory, file_name), "a") as f:
                f.write(intro + report + "\n\n")


def _handle_inspect_callback(inspect_callback, sl, iglobal):
    if _number_of_arguments(inspect_callback) == 1:
        inp = (sl, )
    elif _number_of_arguments(inspect_callback) == 2:
        inp = (sl, iglobal)
    inspect_callback(*inp)


def _handle_terminate_callback(terminate_callback, iglobal, comm):
    terminate = terminate_callback(iglobal)
    check_MPI_equality(terminate, comm(iglobal))
    return terminate


def _MPI_master(comm):
    return get_MPI_params_from_comm(comm)[2]


def _make_callable(obj):
    if callable(obj) and not (isinstance(obj, Minimizer) or isinstance(obj, IterationController) or
                              isinstance(obj, Operator)):
        return obj
    else:
        return lambda x: obj


def _want_metric(mini):
    from .descent_minimizers import L_BFGS, VL_BFGS, SteepestDescent

    # TODO Make this a property of the minimizer?
    if isinstance(mini, (SteepestDescent, L_BFGS, VL_BFGS)):
        return False
    return True


def _number_of_arguments(func):
    from inspect import signature
    return len(signature(func).parameters)


def _is_subdomain(sub_domain, total_domain):
    if not isinstance(sub_domain, (MultiDomain, DomainTuple)):
        raise TypeError
    if isinstance(sub_domain, DomainTuple):
        return sub_domain == total_domain
    return all(kk in total_domain.keys() and vv == total_domain[kk]
               for kk, vv in sub_domain.items())


class _InitializeOperator(Operator):
    def __init__(self, domain, target, std):
        self._domain = MultiDomain.make(domain)
        self._target = MultiDomain.make(target)
        self._has_been_called = False
        self._std = float(std)

    def apply(self, x):
        from ..sugar import from_random, is_linearization, makeDomain
        from ..utilities import myassert

        self._check_input(x)
        if self._has_been_called:
            raise RuntimeError("This operator can only be called once since it returns random numbers.")
        self._has_been_called = True
        if is_linearization(x):
            raise TypeError("Only MultiFields are supported as input")
        diff_keys = set(self._target.keys()) - set(self._domain.keys())
        diff_domain = makeDomain({kk: self._target[kk] for kk in diff_keys})
        diff_fld = from_random(diff_domain, std=self._std)
        myassert(len(set(diff_fld.keys()) & set(self._domain.keys())) == 0)
        return x.unite(diff_fld)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from math import ceil
import sys
from typing import Callable, Iterable, Literal, Optional, Tuple, Union
from warnings import warn

import jax
from jax import numpy as jnp
import numpy as np
from scipy.spatial import distance_matrix

from .forest_util import zeros_like


def get_refinement_shapewithdtype(
    shape0: Union[int, tuple],
    depth: int,
    dtype=None,
    *,
    _coarse_size: int = 3,
    _fine_size: int = 2,
    _fine_strategy: Literal["jump", "extend"] = "jump",
    skip0: bool = False,
):
    from .forest_util import ShapeWithDtype

    if depth < 0:
        raise ValueError(f"invalid `depth`; got {depth!r}")
    csz = int(_coarse_size)  # coarse size
    fsz = int(_fine_size)  # fine size

    swd = partial(ShapeWithDtype, dtype=dtype)

    shape0 = (shape0, ) if isinstance(shape0, int) else shape0
    ndim = len(shape0)
    exc_shp = [swd(shape0)] if not skip0 else [None]
    if depth > 0:
        if _fine_strategy == "jump":
            exc_lvl = tuple(el - (csz - 1) for el in shape0) + (fsz**ndim, )
        elif _fine_strategy == "extend":
            exc_lvl = tuple(
                ceil((el - (csz - 1)) / (fsz // 2)) for el in shape0
            ) + (fsz**ndim, )
        else:
            raise ValueError(f"invalid `_fine_strategy`; got {_fine_strategy}")
        exc_shp += [swd(exc_lvl)]
    for lvl in range(1, depth):
        if _fine_strategy == "jump":
            exc_lvl = tuple(
                fsz * el - (csz - 1) for el in exc_shp[-1].shape[:-1]
            ) + (fsz**ndim, )
        elif _fine_strategy == "extend":
            exc_lvl = tuple(
                ceil((fsz * el - (csz - 1)) / (fsz // 2))
                for el in exc_shp[-1].shape[:-1]
            ) + (fsz**ndim, )
        else:
            raise AssertionError()
        if any(el <= 0 for el in exc_lvl):
            ve = (
                f"`shape0` ({shape0}) with `depth` ({depth}) yield an"
                f" invalid shape ({exc_lvl}) at level {lvl}"
            )
            raise ValueError(ve)
        exc_shp += [swd(exc_lvl)]

    return exc_shp


def coarse2fine_shape(
    shape0: Union[int, Iterable[int]],
    depth: int,
    *,
    _coarse_size: int = 3,
    _fine_size: int = 2,
    _fine_strategy: Literal["jump", "extend"] = "jump",
):
    """Translates a coarse shape to its corresponding fine shape."""
    shape0 = (shape0, ) if isinstance(shape0, int) else shape0
    csz = int(_coarse_size)  # coarse size
    fsz = int(_fine_size)  # fine size
    if _fine_size % 2 != 0:
        raise ValueError("only even numbers allowed for `_fine_size`")

    shape = []
    for shp in shape0:
        sz_at = shp
        for lvl in range(depth):
            if _fine_strategy == "jump":
                sz_at = fsz * (sz_at - (csz - 1))
            elif _fine_strategy == "extend":
                sz_at = fsz * ceil((sz_at - (csz - 1)) / (fsz // 2))
            else:
                ve = f"invalid `_fine_strategy`; got {_fine_strategy}"
                raise ValueError(ve)
            if sz_at <= 0:
                ve = (
                    f"`shape0` ({shape0}) with `depth` ({depth}) yield an"
                    f" invalid shape ({sz_at}) at level {lvl}"
                )
                raise ValueError(ve)
        shape.append(int(sz_at))
    return tuple(shape)


def fine2coarse_shape(
    shape: Union[int, Iterable[int]],
    depth: int,
    *,
    _coarse_size: int = 3,
    _fine_size: int = 2,
    _fine_strategy: Literal["jump", "extend"] = "jump",
    ceil_sizes: bool = False,
):
    """Translates a fine shape to its corresponding coarse shape."""
    shape = (shape, ) if isinstance(shape, int) else shape
    csz = int(_coarse_size)  # coarse size
    fsz = int(_fine_size)  # fine size
    if _fine_size % 2 != 0:
        raise ValueError("only even numbers allowed for `_fine_size`")

    shape0 = []
    for shp in shape:
        sz_at = shp
        for lvl in range(depth, 0, -1):
            if _fine_strategy == "jump":
                # solve for n: `fsz * (n - (csz - 1))`
                sz_at = sz_at / fsz + (csz - 1)
            elif _fine_strategy == "extend":
                # solve for n: `fsz * ceil((n - (csz - 1)) / (fsz // 2))`
                # NOTE, not unique because of `ceil`; use lower limit
                sz_at_max = (sz_at / fsz) * (fsz // 2) + (csz - 1)
                sz_at_min = ceil(sz_at_max - (fsz // 2 - 1))
                for sz_at_cand in range(sz_at_min, ceil(sz_at_max) + 1):
                    try:
                        shp_cand = coarse2fine_shape(
                            (sz_at_cand, ),
                            depth=depth - lvl + 1,
                            _coarse_size=csz,
                            _fine_size=fsz,
                            _fine_strategy=_fine_strategy
                        )[0]
                    except ValueError as e:
                        if "invalid shape" not in "".join(e.args):
                            ve = "unexpected behavior of `coarse2fine_shape`"
                            raise ValueError(ve) from e
                        shp_cand = -1
                    if shp_cand >= shp:
                        sz_at = sz_at_cand
                        break
                else:
                    ve = f"interval search within [{sz_at_min}, {ceil(sz_at_max)}] failed"
                    raise ValueError(ve)
            else:
                ve = f"invalid `_fine_strategy`; got {_fine_strategy}"
                raise ValueError(ve)

            sz_at = ceil(sz_at) if ceil_sizes else sz_at
            if sz_at != int(sz_at):
                raise ValueError(f"invalid shape at level {lvl}")
        shape0.append(int(sz_at))
    return tuple(shape0)


def coarse2fine_distances(
    distances0: Union[float, Iterable[float]],
    depth: int,
    *,
    _fine_size: int = 2,
    _fine_strategy: Literal["jump", "extend"] = "jump",
):
    """Translates coarse distances to its corresponding fine distances."""
    fsz = int(_fine_size)  # fine size
    if _fine_strategy == "jump":
        fpx_in_cpx = fsz**depth
    elif _fine_strategy == "extend":
        fpx_in_cpx = 2**depth
    else:
        ve = f"invalid `_fine_strategy`; got {_fine_strategy}"
        raise ValueError(ve)

    return jnp.atleast_1d(distances0) / fpx_in_cpx


def fine2coarse_distances(
    distances: Union[float, Iterable[float]],
    depth: int,
    *,
    _fine_size: int = 2,
    _fine_strategy: Literal["jump", "extend"] = "jump",
):
    """Translates fine distances to its corresponding coarse distances."""
    fsz = int(_fine_size)  # fine size
    if _fine_strategy == "jump":
        fpx_in_cpx = fsz**depth
    elif _fine_strategy == "extend":
        fpx_in_cpx = 2**depth
    else:
        ve = f"invalid `_fine_strategy`; got {_fine_strategy}"
        raise ValueError(ve)

    return jnp.atleast_1d(distances) * fpx_in_cpx


def _clipping_posdef_logdet(mat, msg_prefix=""):
    sign, logdet = jnp.linalg.slogdet(mat)
    if sign <= 0:
        ve = "not positive definite; clipping eigenvalues"
        warn(msg_prefix + ve)
        eps = jnp.finfo(mat.dtype.type).eps
        evs = jnp.linalg.eigvalsh(mat)
        logdet = jnp.sum(jnp.log(jnp.clip(evs, a_min=eps * evs.max())))
    return logdet


def gauss_kl(cov_desired, cov_approx, *, m_desired=None, m_approx=None):
    cov_t_dl = _clipping_posdef_logdet(cov_desired, msg_prefix="`cov_desired` ")
    cov_a_dl = _clipping_posdef_logdet(cov_approx, msg_prefix="`cov_approx` ")
    cov_a_inv = jnp.linalg.inv(cov_approx)

    kl = -cov_desired.shape[0]  # number of dimensions
    kl += cov_a_dl - cov_t_dl + jnp.trace(cov_a_inv @ cov_desired)
    if m_approx is not None and m_desired is not None:
        m_diff = m_approx - m_desired
        kl += m_diff @ cov_a_inv @ m_diff
    elif not (m_approx is None and m_approx is None):
        ve = "either both or neither of `m_approx` and `m_desired` must be `None`"
        raise ValueError(ve)
    return 0.5 * kl


def refinement_covariance(chart, kernel, jit=True):
    """Computes the implied covariance as modeled by the refinement scheme."""
    from .refine_chart import RefinementField

    cf = RefinementField(chart, kernel=kernel)
    try:
        cf_T = jax.linear_transpose(cf, cf.shapewithdtype)
        cov_implicit = lambda x: cf(*cf_T(x))
        cov_implicit = jax.jit(cov_implicit) if jit else cov_implicit
        _ = cov_implicit(jnp.zeros(chart.shape))  # Test transpose
    except (NotImplementedError, AssertionError):
        # Workaround JAX not yet implementing the transpose of the scanned
        # refinement
        _, cf_T = jax.vjp(cf, zeros_like(cf.shapewithdtype))
        cov_implicit = lambda x: cf(*cf_T(x))
        cov_implicit = jax.jit(cov_implicit) if jit else cov_implicit

    probe = jnp.zeros(chart.shape)
    indices = np.indices(chart.shape).reshape(chart.ndim, -1)
    cov_empirical = jax.lax.map(
        lambda idx: cov_implicit(probe.at[tuple(idx)].set(1.)).ravel(),
        indices.T
    ).T  # vmap over `indices` w/ `in_axes=1, out_axes=-1`

    return cov_empirical


def true_covariance(chart, kernel, depth=None):
    """Computes the true covariance at the final grid."""
    depth = chart.depth if depth is None else depth

    c0_slc = tuple(slice(sz) for sz in chart.shape_at(depth))
    pos = jnp.stack(chart.ind2cart(jnp.mgrid[c0_slc], depth),
                    axis=-1).reshape(-1, chart.ndim)
    dist_mat = distance_matrix(pos, pos)
    return kernel(dist_mat)


def refinement_approximation_error(
    chart,
    kernel: Callable,
    cutout: Optional[Union[slice, int, Tuple[slice], Tuple[int]]] = None,
):
    """Computes the Kullback-Leibler (KL) divergence of the true covariance versus the
    approximative one for a given kernel and shape of the fine grid.

    If the desired shape can not be matched, the next larger one is used and
    the field is subsequently cropped to the desired shape.
    """

    suggested_min_shape = 2 * 4**chart.depth
    if any(s <= suggested_min_shape for s in chart.shape):
        msg = (
            f"shape {chart.shape} potentially too small"
            f" (desired {(suggested_min_shape, ) * chart.ndim} (=`2*4^depth`))"
        )
        warn(msg)

    cov_empirical = refinement_covariance(chart, kernel)
    cov_truth = true_covariance(chart, kernel)

    if cutout is None and all(s > suggested_min_shape for s in chart.shape):
        cutout = (suggested_min_shape, ) * chart.ndim
        print(
            f"cropping field (w/ shape {chart.shape}) to {cutout}",
            file=sys.stderr
        )
    if cutout is not None:
        if isinstance(cutout, slice):
            cutout = (cutout, ) * chart.ndim
        elif isinstance(cutout, int):
            cutout = (slice(cutout), ) * chart.ndim
        elif isinstance(cutout, tuple):
            if all(isinstance(el, slice) for el in cutout):
                pass
            elif all(isinstance(el, int) for el in cutout):
                cutout = tuple(slice(el) for el in cutout)
            else:
                raise TypeError("elements of `cutout` of invalid type")
        else:
            raise TypeError("`cutout` of invalid type")

        cov_empirical = cov_empirical.reshape(chart.shape * 2)[cutout * 2]
        cov_truth = cov_truth.reshape(chart.shape * 2)[cutout * 2]
        sz = np.prod(cov_empirical.shape[:chart.ndim])
        if np.prod(cov_truth.shape[:chart.ndim]) != sz or not sz.dtype == int:
            raise AssertionError()
        cov_empirical = cov_empirical.reshape(sz, sz)
        cov_truth = cov_truth.reshape(sz, sz)

    aux = {
        "cov_empirical": cov_empirical,
        "cov_truth": cov_truth,
    }
    return gauss_kl(cov_truth, cov_empirical), aux



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np
from scipy.interpolate import CubicSpline
from scipy.stats import gamma, invgamma, laplace, norm

from .. import random
from ..domain_tuple import DomainTuple
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from ..operators.adder import Adder
from ..operators.operator import Operator
from ..sugar import makeOp


def _f_on_np(f, arr):
    fld = Field.from_raw(UnstructuredDomain(arr.shape), arr)
    return f(fld).val


class _InterpolationOperator(Operator):
    """
    Calculates a function pointwise on a field by interpolation.
    Can be supplied with a table_func to increase the interpolation accuracy,
    Best results are achieved when table_func(func) is roughly linear.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    func : function
        The function which is applied on the field. Assumed to act on numpy
        arrays.
    xmin : float
        The smallest value for which func will be evaluated.
    xmax : float
        The largest value for which func will be evaluated.
    delta : float
        Distance between sampling points for linear interpolation.
    table_func : function
        Non-linear function applied to table in order to transform the table
        to a more linear space. Assumed to act on `Linearization`s, optional.
    inv_table_func : function
        Inverse of table_func, optional.
    """
    def __init__(self, domain, func, xmin, xmax, delta, table_func=None, inv_table_func=None):
        self._domain = self._target = DomainTuple.make(domain)
        self._xmin, self._xmax = float(xmin), float(xmax)
        self._d = float(delta)
        self._xs = np.arange(xmin, xmax, self._d)
        self._table = func(self._xs)
        if table_func is not None:
            if inv_table_func is None:
                raise ValueError
# MR FIXME: not sure whether we should have this in production code
            a = func(random.current_rng().random(10))
            a1 = _f_on_np(lambda x: inv_table_func(table_func(x)), a)
            np.testing.assert_allclose(a, a1)
            self._table = _f_on_np(table_func, self._table)
        self._interpolator = CubicSpline(self._xs, self._table)
        self._deriv = self._interpolator.derivative()
        self._inv_table_func = inv_table_func

        try:
            from jax import numpy as jnp

            def jax_expr(x):
                res = jnp.interp(x, self._xs, self._table)
                if inv_table_func is not None:
                    ve = (
                        "can not translate arbitrary inverse"
                        f" table function {inv_table_func!r}"
                    )
                    raise ValueError(ve)
                return res

            self._jax_expr = jax_expr
        except ImportError:
            self._jax_expr = None

    def apply(self, x):
        self._check_input(x)
        lin = x.jac is not None
        xval = x.val.val if lin else x.val
        res = self._interpolator(xval)
        res = Field(self._domain, res)
        if lin:
            res = x.new(res, makeOp(Field(self._domain, self._deriv(xval))))
        if self._inv_table_func is not None:
            res = self._inv_table_func(res)
        return res


class InverseGammaOperator(Operator):
    """Transform a standard normal into an inverse gamma distribution.

    The pdf of the inverse gamma distribution is defined as follows:

    .. math::
        \\frac{q^\\alpha}{\\Gamma(\\alpha)}x^{-\\alpha -1}
        \\exp \\left(-\\frac{q}{x}\\right)

    That means that for large x the pdf falls off like :math:`x^{(-\\alpha -1)}`.
    The mean of the pdf is at :math:`q / (\\alpha - 1)` if :math:`\\alpha > 1`.
    The mode is :math:`q / (\\alpha + 1)`.

    The operator can be initialized by setting either alpha and q or mode and mean.
    In accordance to the statements above the mean must be greater
    than the mode. Otherwise would get alpha < 0 and so no mean would be defined.

    This transformation is implemented as a linear interpolation which maps a
    Gaussian onto an inverse gamma distribution.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    alpha : float
        The alpha-parameter of the inverse-gamma distribution.
    q : float or :class:`nifty8.field.Field`
        The q-parameter of the inverse-gamma distribution.
    mode: float
        The mode of the inverse-gamma distribution.
    mean: float
        The mean of the inverse-gamma distribution.
    delta : float
        Distance between sampling points for linear interpolation.
    """
    def __init__(self, domain, alpha=None, q=None, delta=1e-2, mode=None, mean=None):
        self._domain = self._target = DomainTuple.make(domain)

        if alpha is not None and q is not None:
            self._alpha = float(alpha)
            self._q = q if isinstance(q, Field) else float(q)
            self._mode = self._q / (self._alpha + 1)
            if self._alpha > 1:
                self._mean = self._q / (self._alpha - 1)
        elif mean is not None and mode is not None:
            if mean < mode:
                raise ValueError('Mean should be greater than mode, otherwise alpha < 0')
            self._mean = float(mean)
            self._mode = float(mode)
            self._alpha = 2 / (self._mean / self._mode - 1) + 1
            self._q = self._mode * (self._alpha + 1)
        else:
            raise ValueError("Either one pair of arguments (mode, mean or alpha, q) must be given.")

        self._delta = float(delta)
        op = _InterpolationOperator(self._domain, lambda x: invgamma.ppf(norm._cdf(x), float(self._alpha)),
                                    -8.2, 8.2, self._delta, lambda x: x.ptw("log"), lambda x: x.ptw("exp"))
        if np.isscalar(self._q):
            op = op.scale(self._q)
        else:
            op = makeOp(self._q) @ op
        self._op = op

        try:
            from ..re.stats_distributions import invgamma_prior

            q_val = self._q.val if isinstance(self._q, Field) else self._q
            self._jax_expr = invgamma_prior(float(self._alpha), q_val)
        except ImportError:
            self._jax_expr = None

    def apply(self, x):
        return self._op(x)

    @property
    def alpha(self):
        """float : The value of the alpha-parameter of the inverse-gamma distribution"""
        return self._alpha

    @property
    def q(self):
        """float : The value of q-parameters of the inverse-gamma distribution"""
        return self._q

    @property
    def mode(self):
        """float : The value of the mode of the inverse-gamma distribution"""
        return self._mode

    @property
    def mean(self):
        """float : The value of the mean of the inverse-gamma distribution. Only existing for alpha > 1."""
        if self._alpha <= 1:
            raise ValueError('mean only existing for alpha > 1')
        return self._mean

    @property
    def var(self):
        """float : The value of the variance of the inverse-gamma distribution. Only existing for alpha > 2."""
        if self._alpha <= 2:
            raise ValueError('variance only existing for alpha > 2')
        return self._q**2 / ((self._alpha - 1)**2 * (self._alpha - 2))


class GammaOperator(Operator):
    """Transform a standard normal into a gamma distribution.

    The pdf of the gamma distribution is defined as follows:

    .. math::
        \\frac{1}{\\Gamma(k)\\theta^k} x^{k-1}
        \\exp \\left( -\\frac{x}{\\theta} \\right)

    This transformation is implemented as a linear interpolation which maps a
    Gaussian onto a gamma distribution.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    alpha : float
        The shape parameter of the gamma distribution.
    beta : float or :class:`nifty8.field.Field`
        The rate parameter of the gamma distribution.
    theta : float or :class:`nifty8.field.Field`
        The scale parameter of the gamma distribution.
    mean : float
        Mean of the gamma distribution.
    var : float
        Variance of the gamma distribution.
    delta : float
        Distance between sampling points for linear interpolation.
    """
    def __init__(self, domain, alpha=None, beta=None, theta=None, mean=None, var=None, delta=1e-2):
        self._domain = self._target = DomainTuple.make(domain)

        if mean is not None and var is not None:
            theta = var / mean
            alpha = mean / theta

        if beta is None and theta is None:
            raise ValueError("Either beta or theta need to be defined")
        if beta is not None:
            theta = 1 / beta
        self._alpha = float(alpha)
        self._theta = theta if isinstance(theta, Field) else float(theta)

        self._delta = float(delta)
        op = _InterpolationOperator(self._domain, lambda x: gamma.ppf(norm._cdf(x), self._alpha),
                                    -8.2, 8.2, self._delta)
        if np.isscalar(self._theta):
            op = op.scale(self._theta)
        else:
            op = makeOp(self._theta) @ op
        self._op = op

    def apply(self, x):
        return self._op(x)

    @property
    def alpha(self):
        """float : The value of the shape-parameter of the gamma distribution"""
        return self._alpha

    @property
    def beta(self):
        """float : The value of rate parameter of the gamma distribution"""
        return 1 / self._theta

    @property
    def theta(self):
        """float : The value of scale parameter of the gamma distribution"""
        return self._theta

    @property
    def mode(self):
        """float : The value of the mode of the gamma distribution. Exists for alpha >= 1 or k >= 1."""
        if self._alpha < 1:
            raise ValueError('mode only existing for alpha >= 1 or k >= 1')
        return (self._alpha - 1) * self._theta

    @property
    def mean(self):
        """float : The value of the mean of the gamma distribution."""
        return self._alpha * self._theta

    @property
    def var(self):
        """float : The value of the variance of the gamma distribution."""
        return self._alpha * self._theta**2


def LogInverseGammaOperator(domain, alpha, q, delta=1e-2):
    """Transform a standard normal into the log of an inverse gamma distribution.

    See also
    --------
    :class:`InverseGammaOperator`
    """
    op = _InterpolationOperator(domain, lambda x: np.log(invgamma.ppf(norm._cdf(x), float(alpha))),
                                -8.2, 8.2, delta)
    q = np.log(q) if np.isscalar(q) else q.log()
    return Adder(q, domain=op.target) @ op


class UniformOperator(Operator):
    """Transform a standard normal into a uniform distribution.

    The uniform distribution's support is ``[loc, loc + scale]``.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    loc: float
    scale: float

    """
    def __init__(self, domain, loc=0, scale=1):
        self._target = self._domain = DomainTuple.make(domain)
        self._loc = float(loc)
        self._scale = float(scale)

    def apply(self, x):
        self._check_input(x)
        lin = x.jac is not None
        xval = x.val.val if lin else x.val
        res = Field(self._target, self._scale*norm._cdf(xval) + self._loc)
        if not lin:
            return res
        jac = makeOp(Field(self._domain, norm._pdf(xval)*self._scale))
        return x.new(res, jac)

    def inverse(self, field):
        res = norm._ppf((field.val - self._loc) / self._scale)
        return Field(field.domain, res)


class LaplaceOperator(Operator):
    """Transform a standard normal to a Laplace distribution.

    Parameters
    -----------
    domain : Domain, tuple of Domain or DomainTuple
        The domain on which the field shall be defined. This is at the same
        time the domain and the target of the operator.
    loc : float

    scale : float
    """
    def __init__(self, domain, loc=0, scale=1):
        self._target = self._domain = DomainTuple.make(domain)
        self._loc = float(loc)
        self._scale = float(scale)

    def apply(self, x):
        self._check_input(x)
        lin = x.jac is not None
        xval = x.val.val if lin else x.val
        res = Field(self._target, laplace.ppf(norm._cdf(xval), self._loc, self._scale))
        if not lin:
            return res
        y = norm._cdf(xval)
        y = self._scale * np.where(y > 0.5, 1/(1-y), 1/y)
        jac = makeOp(Field(self.domain, y*norm._pdf(xval)))
        return x.new(res, jac)

    def inverse(self, x):
        res = laplace.cdf(x.val, self._loc, self._scale)
        res = norm._ppf(res)
        return Field(x.domain, res)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..logger import logger
from ..minimization.conjugate_gradient import ConjugateGradient
from ..minimization.iteration_controllers import IterationController
from ..minimization.quadratic_energy import QuadraticEnergy
from ..sugar import full
from .endomorphic_operator import EndomorphicOperator
from .linear_operator import LinearOperator


class InversionEnabler(EndomorphicOperator):
    """Class which augments the capability of another operator object via
    numerical inversion.

    Parameters
    ----------
    op : :class:`EndomorphicOperator`
        The operator to be enhanced.
        The InversionEnabler object will support the same operation modes as
        `op`, and additionally the inverse set. The newly-added modes will
        be computed by iterative inversion.
    iteration_controller : :class:`IterationController`
        The iteration controller to use for the iterative numerical inversion
        done by a :class:`ConjugateGradient` object.
    approximation : :class:`LinearOperator`, optional
        if not None, this operator should be an approximation to `op`, which
        supports the operation modes that `op` doesn't have. It is used as a
        preconditioner during the iterative inversion, to accelerate
        convergence.
    """

    def __init__(self, op, iteration_controller, approximation=None):
        # isinstance(op, EndomorphicOperator) does not suffice since op can be
        # a ChainOperator
        if not isinstance(op, LinearOperator):
            raise TypeError('Operator needs to be linear.')
        if op.domain is not op.target:
            raise TypeError('Operator needs to be endomorphic.')
        self._op = op
        self._ic = iteration_controller
        self._approximation = approximation
        self._domain = op.domain
        self._capability = self._addInverse[self._op.capability]

    def apply(self, x, mode):
        self._check_mode(mode)
        if self._op.capability & mode:
            return self._op.apply(x, mode)

        x0 = full(x.domain, 0.)
        invmode = self._modeTable[self.INVERSE_BIT][self._ilog[mode]]
        invop = self._op._flip_modes(self._ilog[invmode])
        prec = self._approximation
        if prec is not None:
            prec = prec._flip_modes(self._ilog[mode])
        energy = QuadraticEnergy(x0, invop, x)
        inverter = ConjugateGradient(self._ic)
        r, stat = inverter(energy, preconditioner=prec)
        if stat != IterationController.CONVERGED:
            logger.warning("Error detected during operator inversion")
        return r.position

    def draw_sample(self, from_inverse=False):
        return self._op.draw_sample(from_inverse)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..utilities import NiftyMeta


class Minimizer(metaclass=NiftyMeta):
    """A base class used by all minimizers."""

    def __call__(self, energy, preconditioner=None):
        """Performs the minimization of the provided Energy functional.

        Parameters
        ----------
        energy : Energy
           Energy object at the starting point of the iteration

        preconditioner : LinearOperator, optional
           Preconditioner to accelerate the minimization

        Returns
        -------
        Energy : Latest `energy` of the minimization.
        int : exit status of the minimization
            Can be controller.CONVERGED or controller.ERROR
        """
        raise NotImplementedError

    @property
    def controller(self):
        raise NotImplementedError



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from ..linearization import Linearization
from ..minimization.energy_adapter import StochasticEnergyAdapter
from ..multi_field import MultiField
from ..operators.einsum import MultiLinearEinsum
from ..operators.energy_operators import EnergyOperator
from ..operators.linear_operator import LinearOperator
from ..operators.multifield2vector import Multifield2Vector
from ..operators.sandwich_operator import SandwichOperator
from ..operators.simple_linear_operators import FieldAdapter
from ..sugar import from_random, full, is_fieldlike, makeDomain, makeField
from ..utilities import myassert


class MeanFieldVI:
    """Collect the operators required for Gaussian meanfield variational
    inference.

    Gaussian meanfield variational inference approximates some target
    distribution with a Gaussian distribution with a diagonal covariance
    matrix. The parameters of the approximation, in this case the mean and
    standard deviation, are obtained by minimizing a stochastic estimate of the
    Kullback-Leibler divergence between the target and the approximation.  In
    order to obtain gradients w.r.t the parameters, the reparametrization trick
    is employed, which separates the stochastic part of the approximation from
    a deterministic function, the generator. Samples from the approximation are
    drawn by processing samples from a standard Gaussian through this
    generator.

    Parameters
    ----------
    position : :class:`nifty8.field.Field`
        The initial estimate of the approximate mean parameter.
    hamiltonian : Energy
        Hamiltonian of the approximated probability distribution.
    n_samples : int
        Number of samples used to stochastically estimate the KL.
    mirror_samples : bool
        Whether the negative of the drawn samples are also used, as they are
        equally legitimate samples. If true, the number of used samples
        doubles. Mirroring samples stabilizes the KL estimate as extreme sample
        variation is counterbalanced. Since it improves stability in many
        cases, it is recommended to set `mirror_samples` to `True`.
    initial_sig : positive :class:`nifty8.field.Field` or positive float
        The initial estimate of the standard deviation.
    comm : MPI communicator or None
        If not None, samples will be distributed as evenly as possible across
        this communicator. If `mirror_samples` is set, then a sample and its
        mirror image will always reside on the same task.
    nanisinf : bool
        If true, nan energies which can happen due to overflows in the forward
        model are interpreted as inf. Thereby, the code does not crash on these
        occasions but rather the minimizer is told that the position it has
        tried is not sensible.
    """
    def __init__(self, position, hamiltonian, n_samples, mirror_samples,
                 initial_sig=1, comm=None, nanisinf=False):
        Flat = Multifield2Vector(position.domain)
        self._std = FieldAdapter(Flat.target, 'std').absolute()
        latent = FieldAdapter(Flat.target,'latent')
        self._mean = FieldAdapter(Flat.target, 'mean')
        self._generator = Flat.adjoint(self._mean + self._std * latent)
        self._entropy = GaussianEntropy(self._std.target) @ self._std
        self._mean = Flat.adjoint @ self._mean
        self._std = Flat.adjoint @ self._std
        pos = {'mean': Flat(position)}
        if is_fieldlike(initial_sig):
            pos['std'] = Flat(initial_sig)
        else:
            pos['std'] = full(Flat.target, initial_sig)
        pos = MultiField.from_dict(pos)
        op = hamiltonian(self._generator) + self._entropy
        self._KL = StochasticEnergyAdapter.make(pos, op, ['latent',], n_samples,
                                    mirror_samples, nanisinf=nanisinf, comm=comm)
        self._samdom = latent.domain

    @property
    def mean(self):
        return self._mean.force(self._KL.position)

    @property
    def std(self):
        return self._std.force(self._KL.position)

    @property
    def entropy(self):
        return self._entropy.force(self._KL.position)

    @property
    def KL(self):
        return self._KL

    def draw_sample(self):
        _, op = self._generator.simplify_for_constant_input(
                from_random(self._samdom))
        return op(self._KL.position)

    def minimize(self, minimizer):
        self._KL, _ = minimizer(self._KL)


class FullCovarianceVI:
    """Collect the operators required for Gaussian full-covariance variational

    Gaussian meanfield variational inference approximates some target
    distribution with a Gaussian distribution with a diagonal covariance
    matrix. The parameters of the approximation, in this case the mean and a
    lower triangular matrix corresponding to a Cholesky decomposition of the
    covariance, are obtained by minimizing a stochastic estimate of the
    Kullback-Leibler divergence between the target and the approximation.  In
    order to obtain gradients w.r.t the parameters, the reparametrization trick
    is employed, which separates the stochastic part of the approximation from
    a deterministic function, the generator. Samples from the approximation are
    drawn by processing samples from a standard Gaussian through this
    generator.

    Note that the size of the covariance scales quadratically with the number
    of model parameters.

    Parameters
    ----------
    position : :class:`nifty8.field.Field`
        The initial estimate of the approximate mean parameter.
    hamiltonian : Energy
        Hamiltonian of the approximated probability distribution.
    n_samples : int
        Number of samples used to stochastically estimate the KL.
    mirror_samples : bool
        Whether the negative of the drawn samples are also used, as they are
        equally legitimate samples. If true, the number of used samples
        doubles. Mirroring samples stabilizes the KL estimate as extreme sample
        variation is counterbalanced. Since it improves stability in many
        cases, it is recommended to set `mirror_samples` to `True`.
    initial_sig : positive float
        The initial estimate for the standard deviation. Initially no
        correlation between the parameters is assumed.
    comm : MPI communicator or None
        If not None, samples will be distributed as evenly as possible across
        this communicator. If `mirror_samples` is set, then a sample and its
        mirror image will always reside on the same task.
    nanisinf : bool
        If true, nan energies which can happen due to overflows in the forward
        model are interpreted as inf. Thereby, the code does not crash on these
        occasions but rather the minimizer is told that the position it has
        tried is not sensible.
    """
    def __init__(self, position, hamiltonian, n_samples, mirror_samples,
                initial_sig=1, comm=None, nanisinf=False):
        Flat = Multifield2Vector(position.domain)
        flat_domain = Flat.target[0]
        mat_space = DomainTuple.make((flat_domain,flat_domain))
        lat = FieldAdapter(Flat.target,'latent')
        LT = LowerTriangularInserter(mat_space)
        tri = FieldAdapter(LT.domain, 'cov')
        mean = FieldAdapter(flat_domain,'mean')
        cov = LT @ tri
        matmul_setup = lat.adjoint @ lat + cov.ducktape_left('co')
        MatMult = MultiLinearEinsum(matmul_setup.target,'ij,j->i',
                                    key_order=('co','latent'))

        self._generator = Flat.adjoint @ (mean + MatMult @ matmul_setup)

        diag_cov = (DiagonalSelector(cov.target) @ cov).absolute()
        self._entropy = GaussianEntropy(diag_cov.target) @ diag_cov
        diag_tri = np.diag(np.full(flat_domain.shape[0], initial_sig))
        pos = MultiField.from_dict(
                {'mean': Flat(position),
                 'cov': LT.adjoint(makeField(mat_space, diag_tri))})
        op = hamiltonian(self._generator) + self._entropy
        self._KL = StochasticEnergyAdapter.make(pos, op, ['latent',], n_samples,
                                    mirror_samples, nanisinf=nanisinf, comm=comm)
        self._mean = Flat.adjoint @ mean
        self._samdom = lat.domain

    @property
    def mean(self):
        return self._mean.force(self._KL.position)

    @property
    def entropy(self):
        return self._entropy.force(self._KL.position)

    @property
    def KL(self):
        return self._KL

    def draw_sample(self):
        _, op = self._generator.simplify_for_constant_input(
                from_random(self._samdom))
        return op(self._KL.position)

    def minimize(self, minimizer):
        self._KL, _ = minimizer(self._KL)


class GaussianEntropy(EnergyOperator):
    """Entropy of a Gaussian distribution given the diagonal of a triangular
    decomposition of the covariance.

    As metric a `SandwichOperator` of the Jacobian is used. This is not a
    proper Fisher metric but may be useful for second order minimization.

    Parameters
    ----------
    domain: Domain, DomainTuple, list of Domain
        The domain of the diagonal.
    """

    def __init__(self, domain):
        self._domain = DomainTuple.make(domain)

    def apply(self, x):
        self._check_input(x)
        if isinstance(x, Field):
             if not np.issubdtype(x.dtype, np.floating):
                 raise NotImplementedError("only real fields are allowed")
        if isinstance(x, MultiField):
             for key in x.keys():
                 if not np.issubdtype(x[key].dtype, np.floating):
                     raise NotImplementedError("only real fields are allowed")
        res = (x*x).scale(2*np.pi*np.e).log().sum().scale(-0.5)
        if not isinstance(x, Linearization):
            return res
        if not x.want_metric:
            return res
        return res.add_metric(SandwichOperator.make(res.jac))


class LowerTriangularInserter(LinearOperator):
    """Insert the entries of a lower triangular matrix into a matrix.

    Parameters
    ----------
    target: Domain, DomainTuple, list of Domain
        A two-dimensional domain with NxN entries.
    """

    def __init__(self, target):
        myassert(len(target.shape) == 2)
        myassert(target.shape[0] == target.shape[1])
        self._target = makeDomain(target)
        ndof = (target.shape[0]*(target.shape[0]+1))//2
        self._domain = makeDomain(UnstructuredDomain(ndof))
        self._indices = np.tril_indices(target.shape[0])
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        x = x.val
        if mode == self.TIMES:
            res = np.zeros(self._target.shape)
            res[self._indices] = x
        else:
            res = x[self._indices].reshape(self._domain.shape)
        return makeField(self._tgt(mode), res)


class DiagonalSelector(LinearOperator):
    """Extract the diagonal of a two-dimensional field.

    Parameters
    ----------
    domain: Domain, DomainTuple, list of Domain
        The two-dimensional domain of the input field. Must be of shape NxN.
    """

    def __init__(self, domain):
        myassert(len(domain.shape) == 2)
        myassert(domain.shape[0] == domain.shape[1])
        self._domain = makeDomain(domain)
        self._target = makeDomain(UnstructuredDomain(domain.shape[0]))
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        return makeField(self._tgt(mode), np.diag(x.val))
